import pandas as pd

import pytest


def test_audit_data_all_pass(custom_client, df_ok, common_kwargs):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_ok, **common_kwargs
    )
    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0


def test_audit_data_with_duplicates(
    custom_client, df_with_duplicates_set2, common_kwargs
):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_duplicates_set2, **common_kwargs
    )
    assert not all_pass
    assert len(case_specific_dfs) == 0
    assert len(fail_dfs) == 2
    assert "D001" in fail_dfs
    # The two duplicate rows should be returned
    assert len(fail_dfs["D001"]) == 2
    assert "D002" in fail_dfs
    ## D002 can not be run with duplicates
    assert fail_dfs["D002"] is None


def test_clean_data_with_duplicates(
    custom_client, df_with_duplicates_set2, common_kwargs
):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_duplicates_set2, **common_kwargs
    )
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_with_duplicates_set2,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        agg_dict={"y": "sum"},
        **common_kwargs
    )
    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0
    assert len(cleaned_df) == 3


def test_clean_data_raises_valueerror(
    custom_client, df_with_duplicates_set2, common_kwargs
):
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_duplicates_set2, **common_kwargs
    )
    with pytest.raises(
        ValueError, match="agg_dict must be provided to resolve D001 failure."
    ):
        custom_client.clean_data(
            df=df_with_duplicates_set2,
            fail_dict=fail_dfs,
            case_specific_dict=case_specific_dfs,
            **common_kwargs
        )


def test_audit_data_with_missing_dates(
    custom_client, df_with_missing_dates, common_kwargs
):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_missing_dates, **common_kwargs
    )
    assert not all_pass
    assert len(case_specific_dfs) == 0
    assert len(fail_dfs) == 1
    assert "D002" in fail_dfs
    assert len(fail_dfs["D002"]) == 2  # Two missing dates should be returned


def test_clean_data_with_missing_dates(
    custom_client, df_with_missing_dates, common_kwargs
):
    # First audit to get fail_dfs and case_specific_dfs
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_missing_dates, **common_kwargs
    )
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_with_missing_dates,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        agg_dict={"y": "sum"},
        **common_kwargs
    )
    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0
    assert len(cleaned_df) == 6  # Two missing rows added.
    assert pd.to_datetime("2023-01-02") in pd.to_datetime(cleaned_df["ds"]).values


def test_audit_data_with_duplicates_and_missing_dates(
    custom_client, df_with_duplicates_and_missing_dates, common_kwargs
):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_duplicates_and_missing_dates, **common_kwargs
    )
    assert not all_pass
    assert len(case_specific_dfs) == 0
    assert len(fail_dfs) == 2
    assert "D001" in fail_dfs
    assert len(fail_dfs["D001"]) == 2  # The two duplicate rows should be returned
    assert "D002" in fail_dfs
    assert fail_dfs["D002"] is None  # D002 can not be run with duplicates


def test_clean_data_with_duplicates_and_missing_dates(
    custom_client, df_with_duplicates_and_missing_dates, common_kwargs
):
    # First audit to get fail_dfs and case_specific_dfs
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_duplicates_and_missing_dates, **common_kwargs
    )

    # Clean Data (pass 1 will clear the duplicates)
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_with_duplicates_and_missing_dates,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        agg_dict={"y": "sum"},
        **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 1
    # Since duplicates have been removed, D002 has been run now.
    assert "D002" in fail_dfs
    assert len(fail_dfs["D002"]) == 1
    assert len(case_specific_dfs) == 0
    assert len(cleaned_df) == 4  # Two duplicates rows consolidated into one.

    # Clean Data (pass 2 will clear the missing dates)
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=cleaned_df,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        **common_kwargs
    )
    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0
    # Two duplicates rows consolidated into one plus one missing row added.
    assert len(cleaned_df) == 5


def test_audit_data_with_cat_columns(custom_client, df_with_cat_columns, common_kwargs):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_with_cat_columns, **common_kwargs
    )
    assert not all_pass
    assert len(case_specific_dfs) == 0
    assert len(fail_dfs) == 1
    assert "F001" in fail_dfs
    assert fail_dfs["F001"].shape[1] == 2  # Should return both categorical columns


def test_audit_data_with_negative_vals(custom_client, df_negative_vals, common_kwargs):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_negative_vals, **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 1
    assert "V001" in case_specific_dfs
    assert case_specific_dfs["V001"].shape[0] == 3  # should return all negative values


def test_clean_data_with_negative_vals_without_cleaning_case_specific(
    custom_client, df_negative_vals, common_kwargs
):
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_negative_vals, **common_kwargs
    )
    _, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_negative_vals,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        # clean_case_specific=False, # Default
        **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 1
    assert "V001" in case_specific_dfs
    assert case_specific_dfs["V001"].shape[0] == 3  # should return all negative values


def test_clean_data_with_negative_vals_cleaning_case_specific(
    custom_client, df_negative_vals, common_kwargs
):
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_negative_vals, **common_kwargs
    )
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_negative_vals,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        clean_case_specific=True,
        **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 1
    assert "V002" in case_specific_dfs
    assert case_specific_dfs["V002"].shape[0] == 1  # should return leading zeros

    # test second pass
    # Clean Data, second pass (removes leading zeros)
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=cleaned_df,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        clean_case_specific=True,
        **common_kwargs
    )

    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0


def test_audit_data_leading_zeros(custom_client, common_kwargs, df_leading_zeros_set2):
    all_pass, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_leading_zeros_set2, **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 1
    assert "V002" in case_specific_dfs
    assert (
        case_specific_dfs["V002"].shape[0] == 2
    )  # should return ids with leading zeros


def test_clean_data_leading_zeroes_without_cleaning_case_specific(
    custom_client, common_kwargs, df_leading_zeros_set2
):
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_leading_zeros_set2, **common_kwargs
    )
    _, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_leading_zeros_set2,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        # clean_case_specific=False,  # Default
        **common_kwargs
    )
    assert not all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 1
    assert "V002" in case_specific_dfs
    assert (
        case_specific_dfs["V002"].shape[0] == 2
    )  # should return ids with leading zeros


def test_clean_data_with_cleaning_case_specific(
    custom_client, common_kwargs, df_leading_zeros_set2
):
    _, fail_dfs, case_specific_dfs = custom_client.audit_data(
        df=df_leading_zeros_set2, **common_kwargs
    )
    cleaned_df, all_pass, fail_dfs, case_specific_dfs = custom_client.clean_data(
        df=df_leading_zeros_set2,
        fail_dict=fail_dfs,
        case_specific_dict=case_specific_dfs,
        clean_case_specific=True,
        **common_kwargs
    )
    assert all_pass
    assert len(fail_dfs) == 0
    assert len(case_specific_dfs) == 0
    assert len(cleaned_df) == 7  # all leading zeros removed, zero series unchanged



from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd

from xiuhmolpilli.arena import FoundationalTimeSeriesArena
from .utils import models
from .test_eval import generate_exp_dataset


def generate_data(freq: str, tmpdir: str) -> str:
    df = generate_exp_dataset(n_series=5, freq=freq, return_df=True)
    df_parquet_path = Path(tmpdir) / f"dataset_{freq}.parquet"
    df.to_parquet(df_parquet_path)
    return str(df_parquet_path)


def test_foundational_time_series_arena():
    cwd = Path.cwd()
    with TemporaryDirectory(dir=cwd) as tmpdir:
        parquet_data_paths = [generate_data(freq, tmpdir) for freq in ["H", "MS"]]
        arena = FoundationalTimeSeriesArena(
            models=models,
            parquet_data_paths=parquet_data_paths,
            results_dir=tmpdir,
        )
        arena.compete()
        eval_df = pd.read_csv(arena.evaluation_path)
        arena.compete()
        eval_df_2 = pd.read_csv(arena.evaluation_path)
        print(eval_df)
        print(eval_df_2)
        assert eval_df.equals(eval_df_2)
        print(eval_df)



from .ml import AutoLGBM
from .neural import (
    AutoNHITS,
    AutoTFT,
)
from .prophet import NixtlaProphet
from .stats import (
    ADIDA,
    AutoARIMA,
    AutoCES,
    AutoETS,
    CrostonClassic,
    DOTheta,
    HistoricAverage,
    IMAPA,
    SeasonalNaive,
    Theta,
    ZeroModel,
)

__all__ = [
    "AutoLGBM",
    "NixtlaProphet",
    "AutoNHITS",
    "AutoTFT",
    "ADIDA",
    "AutoARIMA",
    "AutoCES",
    "AutoETS",
    "CrostonClassic",
    "DOTheta",
    "HistoricAverage",
    "IMAPA",
    "SeasonalNaive",
    "Theta",
    "ZeroModel",
]



import os
from time import time
from typing import List, Tuple

os.environ["NIXTLA_NUMBA_RELEASE_GIL"] = "1"
os.environ["NIXTLA_NUMBA_CACHE"] = "1"

import fire
import numpy as np
import pandas as pd
from scipy.stats import norm
from statsforecast import StatsForecast
from statsforecast.models import (
    AutoARIMA,
    AutoETS,
    AutoCES,
    DynamicOptimizedTheta,
    SeasonalNaive,
)

from src.utils import ExperimentHandler


def run_seasonal_naive(
    train_df: pd.DataFrame,
    horizon: int,
    freq: str,
    seasonality: int,
    level: List[int],
) -> Tuple[pd.DataFrame, float, str]:
    os.environ["NIXTLA_ID_AS_COL"] = "true"
    sf = StatsForecast(
        models=[SeasonalNaive(season_length=seasonality)],
        freq=freq,
        n_jobs=-1,
    )
    model = sf
    init_time = time()
    fcsts_df = model.forecast(df=train_df, h=horizon, level=level)
    total_time = time() - init_time
    return fcsts_df, total_time, "SeasonalNaive"


def ensemble_forecasts(
    fcsts_df: pd.DataFrame,
    quantiles: List[float],
    name_models: List[str],
    model_name: str,
) -> pd.DataFrame:
    fcsts_df[model_name] = fcsts_df[name_models].mean(axis=1).values  # type: ignore
    # compute quantiles based on the mean of the forecasts
    sigma_models = []
    for model in name_models:
        fcsts_df[f"sigma_{model}"] = fcsts_df[f"{model}-hi-68.27"] - fcsts_df[model]
        sigma_models.append(f"sigma_{model}")
    fcsts_df[f"std_{model_name}"] = (
        fcsts_df[sigma_models].pow(2).sum(axis=1).div(len(sigma_models) ** 2).pow(0.5)
    )
    z = norm.ppf(quantiles)
    q_cols = []
    for q, zq in zip(quantiles, z):
        q_col = f"{model_name}-q-{q}"
        fcsts_df[q_col] = fcsts_df[model_name] + zq * fcsts_df[f"std_{model_name}"]
        q_cols.append(q_col)
    fcsts_df = fcsts_df[["unique_id", "ds"] + [model_name] + q_cols]
    return fcsts_df


def run_statistical_ensemble(
    train_df: pd.DataFrame,
    horizon: int,
    freq: str,
    seasonality: int,
    quantiles: List[float],
) -> Tuple[pd.DataFrame, float, str]:
    os.environ["NIXTLA_ID_AS_COL"] = "true"
    models = [
        AutoARIMA(season_length=seasonality),
        AutoETS(season_length=seasonality),
        AutoCES(season_length=seasonality),
        DynamicOptimizedTheta(season_length=seasonality),
    ]
    init_time = time()
    series_per_core = 15
    n_series = train_df["unique_id"].nunique()
    n_jobs = min(n_series // series_per_core, os.cpu_count())
    sf = StatsForecast(
        models=models,
        freq=freq,
        n_jobs=n_jobs,
    )
    fcsts_df = sf.forecast(df=train_df, h=horizon, level=[68.27])
    name_models = [repr(model) for model in models]
    model_name = "StatisticalEnsemble"
    fcsts_df = ensemble_forecasts(
        fcsts_df,
        quantiles,
        name_models,
        model_name,
    )
    total_time = time() - init_time
    return fcsts_df, total_time, model_name


def main(dataset: str):
    exp = ExperimentHandler(dataset)
    # seasonal naive benchmark
    fcst_df, total_time, model_name = run_seasonal_naive(
        train_df=exp.train_df,
        horizon=exp.horizon,
        freq=exp.freq,
        seasonality=exp.seasonality,
        level=exp.level,
    )
    fcst_df = exp.fcst_from_level_to_quantiles(fcst_df, model_name)
    exp.save_results(fcst_df, total_time, model_name)
    # statistical ensemble
    fcst_df, total_time, model_name = run_statistical_ensemble(
        train_df=exp.train_df,
        horizon=exp.horizon,
        freq=exp.freq,
        seasonality=exp.seasonality,
        quantiles=exp.quantiles,
    )
    exp.save_results(fcst_df, total_time, model_name)


if __name__ == "__main__":
    from statsforecast.utils import AirPassengers as ap

    AutoARIMA(season_length=12).forecast(ap.astype(np.float32), h=12)
    fire.Fire(main)



import pytest

from nixtla.nixtla_client import ApiError
from nixtla_tests.helpers.checks import check_finetuned_model
from nixtla_tests.helpers.states import model_ids_object
from utilsforecast.evaluation import evaluate
from utilsforecast.losses import rmse


class TestTimeSeriesDataSet1:
    def test_finetuning_and_forecasting(self, custom_client, ts_data_set1):
        # Finetune the model
        finetune_resp = custom_client.finetune(
            ts_data_set1.train, output_model_id=model_ids_object.model_id1
        )
        assert finetune_resp == model_ids_object.model_id1

        model_id2 = custom_client.finetune(
            ts_data_set1.train, finetuned_model_id=model_ids_object.model_id1
        )
        model_ids_object.model_id2 = model_id2  # store the model_id2 for later use

        # Forecast with fine-tuned models
        forecast_base = custom_client.forecast(ts_data_set1.train, h=ts_data_set1.h)
        forecast1 = custom_client.forecast(
            ts_data_set1.train,
            h=ts_data_set1.h,
            finetuned_model_id=model_ids_object.model_id1,
        )
        forecast2 = custom_client.forecast(
            ts_data_set1.train,
            h=ts_data_set1.h,
            finetuned_model_id=model_ids_object.model_id2,
        )
        all_fcsts = forecast_base.assign(
            ten_rounds=forecast1["TimeGPT"], twenty_rounds=forecast2["TimeGPT"]
        )
        fcst_rmse = evaluate(
            all_fcsts.merge(ts_data_set1.valid),
            metrics=[rmse],
            agg_fn="mean",
        ).loc[0]
        # error was reduced over 40% by finetuning
        assert 1 - fcst_rmse["ten_rounds"] / fcst_rmse["TimeGPT"] > 0.4
        # error was reduced over 30% by further finetuning
        assert 1 - fcst_rmse["twenty_rounds"] / fcst_rmse["ten_rounds"] > 0.3

        # non-existent model returns 404
        with pytest.raises(ApiError) as excinfo:
            custom_client.forecast(
                ts_data_set1.train, h=ts_data_set1.h, finetuned_model_id="unexisting"
            )
        assert getattr(excinfo.value, "status_code", None) == 404

        # Enough data to finetune
        _, model_horizon = custom_client._get_model_params("timegpt-1", ts_data_set1.freq)
        forecast3 = custom_client.forecast(ts_data_set1.train.tail(model_horizon + 1), h=ts_data_set1.h, finetune_steps=10)

        # Not enough data to finetune
        with pytest.raises(
            ValueError, match="Some series are too short. Please make sure that each series contains at least 8 observations."
        ):
            custom_client.forecast(ts_data_set1.train.tail(model_horizon), h=ts_data_set1.h, finetune_steps=10)

    def test_cv_with_finetuned_model(self, custom_client, ts_data_set1):
        cv_base = custom_client.cross_validation(
            ts_data_set1.series, n_windows=2, h=ts_data_set1.h
        )
        cv_finetune = custom_client.cross_validation(
            ts_data_set1.series,
            n_windows=2,
            h=ts_data_set1.h,
            finetuned_model_id=model_ids_object.model_id1,
        )
        merged = cv_base.merge(
            cv_finetune,
            on=["unique_id", "ds", "cutoff", "y"],
            suffixes=("_base", "_finetune"),
        ).drop(columns="cutoff")
        cv_rmse = evaluate(
            merged,
            metrics=[rmse],
            agg_fn="mean",
        ).loc[0]
        # error was reduced over 40% by finetuning
        assert 1 - cv_rmse["TimeGPT_finetune"] / cv_rmse["TimeGPT_base"] > 0.4

        custom_client.delete_finetuned_model(model_ids_object.model_id1)

    def test_anomaly_detection_with_finetuned_model(
        self, custom_client, ts_anomaly_data
    ):
        anomaly_base = custom_client.detect_anomalies(ts_anomaly_data.train_anomalies)
        anomaly_finetune = custom_client.detect_anomalies(
            ts_anomaly_data.train_anomalies,
            finetuned_model_id=model_ids_object.model_id2,
        )
        detected_anomalies_base = (
            anomaly_base.set_index("ds")
            .loc[ts_anomaly_data.anomaly_date, "anomaly"]
            .sum()
        )
        detected_anomalies_finetune = (
            anomaly_finetune.set_index("ds")
            .loc[ts_anomaly_data.anomaly_date, "anomaly"]
            .sum()
        )
        assert detected_anomalies_base < detected_anomalies_finetune

    def test_list_finetuned_models(self, custom_client):
        models = custom_client.finetuned_models()
        ids = {m.id for m in models}
        assert (
            model_ids_object.model_id1 not in ids and model_ids_object.model_id2 in ids
        )

    def test_get_single_finetuned_model(self, custom_client):
        single_model = custom_client.finetuned_model(model_ids_object.model_id2)
        assert single_model.id == model_ids_object.model_id2
        assert single_model.base_model_id == model_ids_object.model_id1

    def test_non_existing_model_returns_error(self, custom_client):
        with pytest.raises(ApiError, match="Model not found"):
            custom_client.finetuned_model("hi")

    @pytest.mark.distributed_run
    @pytest.mark.ray_run
    def test_ray_finetune_model(self, custom_client, ray_df):
        check_finetuned_model(custom_client, ray_df, model_ids_object.model_id2)

    @pytest.mark.distributed_run
    @pytest.mark.spark_run
    def test_spark_finetune_model(self, custom_client, spark_df):
        check_finetuned_model(custom_client, spark_df, model_ids_object.model_id2)

    @pytest.mark.distributed_run
    @pytest.mark.flaky(reruns=3, delay=10)
    def test_dask_finetune_model(self, custom_client, dask_df):
        check_finetuned_model(custom_client, dask_df, model_ids_object.model_id2)



from gluonts.torch.model.predictor import PyTorchPredictor
from lag_llama.gluon.estimator import LagLlamaEstimator

from ..utils.gluonts_forecaster import GluonTSForecaster


class LagLlama(GluonTSForecaster):
    def __init__(
        self,
        repo_id: str = "time-series-foundation-models/Lag-Llama",
        filename: str = "lag-llama.ckpt",
        alias: str = "LagLlama",
    ):
        super().__init__(
            repo_id=repo_id,
            filename=filename,
            alias=alias,
        )

    def get_predictor(self, prediction_length: int) -> PyTorchPredictor:
        ckpt = self.load()
        estimator_args = ckpt["hyper_parameters"]["model_kwargs"]
        # this context length is reported in the paper
        context_length = 32
        estimator = LagLlamaEstimator(
            ckpt_path=self.checkpoint_path,
            prediction_length=prediction_length,
            context_length=context_length,
            # estimator args
            input_size=estimator_args["input_size"],
            n_layer=estimator_args["n_layer"],
            n_embd_per_head=estimator_args["n_embd_per_head"],
            n_head=estimator_args["n_head"],
            scaling=estimator_args["scaling"],
            time_feat=estimator_args["time_feat"],
        )
        lightning_module = estimator.create_lightning_module()
        transformation = estimator.create_transformation()
        predictor = estimator.create_predictor(transformation, lightning_module)
        return predictor



import pytest

try:
    from pyspark.sql import SparkSession

    @pytest.fixture(scope="module")
    def spark_client():
        with SparkSession.builder.getOrCreate() as spark:
            yield spark

    @pytest.fixture(scope="module")
    def spark_df(spark_client, distributed_series):
        spark_df = spark_client.createDataFrame(distributed_series).repartition(2)
        return spark_df

    @pytest.fixture(scope="module")
    def spark_diff_cols_df(spark_client, distributed_series, renamer):
        spark_df = spark_client.createDataFrame(
            distributed_series.rename(columns=renamer)
        ).repartition(2)
        return spark_df

    @pytest.fixture(scope="module")
    def spark_df_x(spark_client, distributed_df_x):
        spark_df = spark_client.createDataFrame(distributed_df_x).repartition(2)
        return spark_df

    @pytest.fixture(scope="module")
    def spark_df_x_diff_cols(spark_client, distributed_df_x, renamer):
        spark_df = spark_client.createDataFrame(
            distributed_df_x.rename(columns=renamer)
        ).repartition(2)
        return spark_df

    @pytest.fixture(scope="module")
    def spark_future_ex_vars_df(spark_client, distributed_future_ex_vars_df):
        spark_df = spark_client.createDataFrame(
            distributed_future_ex_vars_df
        ).repartition(2)
        return spark_df

    @pytest.fixture(scope="module")
    def spark_future_ex_vars_df_diff_cols(
        spark_client, distributed_future_ex_vars_df, renamer
    ):
        spark_df = spark_client.createDataFrame(
            distributed_future_ex_vars_df.rename(columns=renamer)
        ).repartition(2)
        return spark_df
except ImportError:
    # If PySpark is not installed, we skip the fixtures
    pytest.skip(
        "PySpark is not installed, skipping Spark fixtures", allow_module_level=True
    )



import itertools
import logging
import time

import pandas as pd
from nixtla import NixtlaClient as V2Client
from nixtlats import NixtlaClient as V1Client
from utilsforecast.data import generate_series
from utilsforecast.feature_engineering import fourier

logging.getLogger("nixtla").setLevel(logging.ERROR)
logging.getLogger("nixtlats").setLevel(logging.ERROR)


def forecast(client, df, X_df, h, level):
    return client.forecast(df=df, X_df=X_df, h=h, level=level)


def cross_validation(client, df, X_df, h, level):
    return client.cross_validation(df=df, h=h, n_windows=4, level=level)


def anomaly_detection(client, df, X_df, h, level):
    if isinstance(level, list):
        level = level[0]
    return client.detect_anomalies(df=df, level=level)


v1_client = V1Client()
v2_client = V2Client()
n_series = 1_000
freq = "D"
h = 14
series = generate_series(n_series, freq=freq, min_length=200)
train, future = fourier(series, freq=freq, season_length=7, k=4, h=h)
features = ["none", "exog"]
level = [None, [80]]
clients = {"v1": v1_client, "v2": v2_client}
methods = {
    "forecast": forecast,
    "cross_validation": cross_validation,
    "anomaly_detection": anomaly_detection,
}
times = {version: {} for version in ("v1", "v2")}
for feats, lvl in itertools.product(features, level):
    if feats == "none":
        df = series
        X_df = None
    else:
        df = train
        X_df = future
    for name, method in methods.items():
        if name == "anomaly_detection" and lvl is None:
            continue
        for version, client in clients.items():
            start = time.perf_counter()
            combination = f"{version} {name}. Features: {feats}. Level: {lvl}"
            print(f"Running {combination}")
            res = method(client, df=df, X_df=X_df, h=h, level=lvl)
            time_taken = time.perf_counter() - start
            times[version][f"{name}-{feats}-{lvl}"] = time_taken
            print(f"{combination} took {time_taken:.1f} seconds.")

df = pd.DataFrame(times)
df.index = df.index.str.split("-", expand=True)
df.index.names = ["endpoint", "features", "level"]
df = df.sort_index()
df["speedup"] = df["v1"] / df["v2"]
df["speedup"] = df["speedup"].map("{:.0f}x".format)
for col in ("v1", "v2"):
    df[col] = df[col].map("{:.0f}s".format)
with open("endpoint_times.md", "wt") as f:
    f.write(df.reset_index().to_markdown(index=False))



import logging
import subprocess
from typing import Literal

import fire
import pandas as pd


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

datasets = [
    "m1_yearly",
    "m1_quarterly",
    "m1_monthly",
    "m3_yearly",
    "m3_quarterly",
    "m3_monthly",
    "m3_other",
    "m4_yearly",
    "m4_quarterly",
    "m4_monthly",
    "m4_weekly",
    "m4_daily",
    "m4_hourly",
    "tourism_yearly",
    "tourism_quarterly",
    "tourism_monthly",
]


def main(mode: Literal["fcst_statsforecast", "fcst_moirai`", "evaluation"]):
    prefix_process = ["python", "-m"]

    if mode in ["fcst_statsforecast", "fcst_moirai"]:
        for dataset in datasets:
            logger.info(f"Forecasting {dataset}...")
            suffix_process = ["--dataset", dataset]

            def process(middle_process):
                return prefix_process + middle_process + suffix_process

            if mode == "fcst_statsforecast":
                logger.info("Running StatisticalEnsemble")
                subprocess.run(process(["src.statsforecast_pipeline"]))
            elif mode == "fcst_moirai":
                logger.info("Running SalesforceMoirai")
                subprocess.run(process(["src.moirai_pipeline"]))
    elif mode == "evaluation":
        from src.utils import ExperimentHandler

        eval_df = []
        for dataset in datasets:
            logger.info(f"Evaluating {dataset}...")
            exp = ExperimentHandler(dataset)
            try:
                eval_dataset_df = exp.evaluate_models(
                    [
                        "SalesforceMoirai",
                        "StatisticalEnsemble",
                        "SeasonalNaive",
                    ]
                )
                print(eval_dataset_df)
                eval_df.append(eval_dataset_df)
            except Exception as e:
                logger.error(e)
        eval_df = pd.concat(eval_df).reset_index(drop=True)
        exp.save_dataframe(eval_df, "complete-results.csv")
    else:
        raise ValueError(f"mode {mode} not found")


if __name__ == "__main__":
    fire.Fire(main)



from typing import Iterable, List

import numpy as np
import pandas as pd
import torch
from chronos import ChronosPipeline
from tqdm import tqdm
from utilsforecast.processing import make_future_dataframe

from ..utils.forecaster import Forecaster


class TimeSeriesDataset:
    def __init__(
        self,
        data: torch.Tensor,
        uids: Iterable,
        last_times: Iterable,
        batch_size: int,
    ):
        self.data = data
        self.uids = uids
        self.last_times = last_times
        self.batch_size = batch_size
        self.n_batches = len(data) // self.batch_size + (
            0 if len(data) % self.batch_size == 0 else 1
        )
        self.current_batch = 0

    @classmethod
    def from_df(cls, df: pd.DataFrame, batch_size: int):
        num_unique_ids = df["unique_id"].nunique()
        max_series_length = df["unique_id"].value_counts().max()
        padded_tensor = torch.full(
            size=(num_unique_ids, max_series_length),
            fill_value=torch.nan,
            dtype=torch.bfloat16,
        )  # type: ignore
        df_sorted = df.sort_values(by=["unique_id", "ds"])
        for idx, (_, group) in enumerate(df_sorted.groupby("unique_id")):
            series_length = len(group)
            padded_tensor[idx, -series_length:] = torch.tensor(
                group["y"].values,
                dtype=torch.bfloat16,
            )
        uids = df_sorted["unique_id"].unique()
        last_times = df_sorted.groupby("unique_id")["ds"].tail(1)
        return cls(padded_tensor, uids, last_times, batch_size)

    def __len__(self):
        return self.n_batches

    def make_future_dataframe(self, h: int, freq: str) -> pd.DataFrame:
        return make_future_dataframe(
            uids=self.uids,
            last_times=pd.to_datetime(self.last_times),
            h=h,
            freq=freq,
        )  # type: ignore

    def __iter__(self):
        self.current_batch = 0  # Reset for new iteration
        return self

    def __next__(self):
        if self.current_batch < self.n_batches:
            start_idx = self.current_batch * self.batch_size
            end_idx = start_idx + self.batch_size
            self.current_batch += 1
            return self.data[start_idx:end_idx]
        else:
            raise StopIteration


class Chronos(Forecaster):
    def __init__(
        self,
        repo_id: str = "amazon/chronos-t5-large",
        batch_size: int = 16,
        alias: str = "Chronos",
    ):
        self.repo_id = repo_id
        self.batch_size = batch_size
        self.alias = alias
        self.model = ChronosPipeline.from_pretrained(
            repo_id,
            device_map="auto",
            torch_dtype=torch.bfloat16,
        )

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        dataset = TimeSeriesDataset.from_df(df, batch_size=self.batch_size)
        fcsts = [
            self.model.predict(batch, prediction_length=h) for batch in tqdm(dataset)
        ]
        fcst = torch.cat(fcsts)
        fcst = fcst.numpy()
        fcst_df = dataset.make_future_dataframe(h=h, freq=freq)
        fcst_df[self.alias] = np.mean(fcst, axis=1).reshape(-1, 1)
        return fcst_df



import logging
import subprocess

import pandas as pd

from src.utils import ExperimentHandler

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

not_included_datasets = [
    "m1_yearly",
    "m1_quarterly",
    "m1_monthly",
    "m3_yearly",
    "m3_quarterly",
    "m3_monthly",
    "m3_other",
    "m4_yearly",
    "m4_quarterly",
    "m4_monthly",
    "m4_weekly",
    "m4_daily",
    "m4_hourly",
    "tourism_yearly",
    "tourism_quarterly",
    "tourism_monthly",
]

test_paper_datasets = [
    "pedestrian_counts",
    "weather",
]

datasets = {
    "not_included": not_included_datasets,
    "test_set": test_paper_datasets,
}


def evaluate():
    eval_df = []
    prefix_process = ["python", "-m"]

    for name_group, groups in datasets.items():
        for dataset in groups:
            logger.info(f"Evaluating {dataset}...")
            suffix_process = ["--dataset", dataset]
            process = (
                lambda middle_process: prefix_process + middle_process + suffix_process
            )
            # running statsforecast and lagllama in separated
            # processes because gluonts sets multiprocessing context
            # see: https://github.com/awslabs/gluonts/blob/dev/src/gluonts/torch/__init__.py
            logger.info("Running SeasonalNaive")
            subprocess.run(process(["src.statsforecast_pipeline"]))
            logger.info("Running LagLLama")
            subprocess.run(process(["src.lag_llama_pipeline"]))
            logger.info("Running dataset evaluation")
            exp = ExperimentHandler(dataset)
            eval_dataset_df = exp.evaluate_models(["LagLlama", "SeasonalNaive"])
            eval_dataset_df.insert(0, "paper", name_group)
            eval_df.append(eval_dataset_df)
    eval_df = pd.concat(eval_df).reset_index(drop=True)
    exp.save_dataframe(eval_df, "complete-results.csv")


if __name__ == "__main__":
    evaluate()



import logging
import warnings
from dataclasses import dataclass, asdict
from functools import partial
from pathlib import Path

import pandas as pd
from utilsforecast.evaluation import evaluate
from utilsforecast.losses import rmse, mae, mase

from src.utils.filter_data import DatasetParams

warnings.simplefilter(action="ignore", category=FutureWarning)
logging.basicConfig(level=logging.INFO)
main_logger = logging.getLogger(__name__)


@dataclass
class ExperimentDataset:
    Y_df_train: pd.DataFrame
    Y_df_test: pd.DataFrame
    horizon: int
    seasonality: int
    frequency: str
    pandas_frequency: str

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "ExperimentDataset":
        """
        Parameters
        ----------
        df : pd.DataFrame
            df should have columns: unique_id, ds, y, frequency, pandas_frequency, horizon, seasonality
        """
        ds_params = DatasetParams.from_df(df)
        df = df[["unique_id", "ds", "y"]]  # type: ignore
        Y_df_test = df.groupby("unique_id").tail(ds_params.horizon)
        Y_df_train = df.drop(Y_df_test.index)  # type: ignore
        return cls(
            Y_df_train=Y_df_train,
            Y_df_test=Y_df_test,
            **asdict(ds_params),
        )

    @classmethod
    def from_parquet(
        cls,
        parquet_path: str,
    ) -> "ExperimentDataset":
        df = pd.read_parquet(parquet_path)
        return cls.from_df(df=df)

    def evaluate_forecast_df(
        self,
        forecast_df: pd.DataFrame,
        model: str,
        total_time: float,
    ) -> pd.DataFrame:
        df_ = self.Y_df_test.copy(deep=True)
        if forecast_df.dtypes["ds"] != df_.dtypes["ds"]:
            df_["ds"] = df_["ds"].astype(forecast_df.dtypes["ds"])
        df = df_.merge(
            forecast_df[["unique_id", "ds", model]],
            on=["unique_id", "ds"],
            how="left",
        )
        if df[model].isna().sum() > 0:
            na_uids = df.loc[df[model].isna()]["unique_id"].unique()
            main_logger.warning(
                f"{model} contains NaN for {len(na_uids)} series: {na_uids}"
                "filling with last values"
            )
            from statsforecast import StatsForecast
            from statsforecast.models import SeasonalNaive

            sf = StatsForecast(
                models=[SeasonalNaive(season_length=self.seasonality)],
                freq=self.pandas_frequency,
            )
            sn_df = sf.forecast(
                df=self.Y_df_train,
                h=self.horizon,
            )
            df = df.merge(sn_df, on=["unique_id", "ds"], how="left")  # type: ignore
            df.loc[df["unique_id"].isin(na_uids), model] = df.loc[
                df["unique_id"].isin(na_uids), "SeasonalNaive"
            ]
            df = df.drop(columns=["SeasonalNaive"])
        partial_mase = partial(mase, seasonality=self.seasonality)
        eval_df = evaluate(
            df=df,
            metrics=[rmse, mae, partial_mase],
            train_df=self.Y_df_train,
            models=[model],
        )
        eval_df = eval_df.groupby("metric").mean(numeric_only=True).reset_index()  # type: ignore
        eval_time_df = pd.DataFrame(
            {
                "metric": ["total_time"],
                model: [total_time],
            }
        )
        eval_df = pd.concat(
            [eval_df, eval_time_df],
            ignore_index=True,
        )  # type: ignore
        return eval_df.set_index("metric")


@dataclass
class ForecastDataset:
    forecast_df: pd.DataFrame
    total_time: float

    @classmethod
    def from_dir(cls, dir: str | Path):
        dir_ = Path(dir)
        forecast_df = pd.read_parquet(dir_ / "forecast_df.parquet")
        with open(dir_ / "total_time.txt", "r") as file:
            total_time = float(file.read())
        return cls(forecast_df=forecast_df, total_time=total_time)

    @staticmethod
    def is_forecast_ready(dir: str | Path):
        dir_ = Path(dir)
        forecast_path = dir_ / "forecast_df.parquet"
        time_path = dir_ / "total_time.txt"
        return forecast_path.exists() and time_path.exists()

    def save_to_dir(self, dir: str | Path):
        dir_ = Path(dir)
        dir_.mkdir(parents=True, exist_ok=True)
        self.forecast_df.to_parquet(dir_ / "forecast_df.parquet")
        with open(dir_ / "total_time.txt", "w") as file:
            file.write(str(self.total_time))



import httpx
import pytest
import time

from itertools import product
from nixtla.nixtla_client import (
    ApiError,
)
from nixtla_tests.helpers.checks import check_retry_behavior


def raise_api_error_with_text(*args, **kwargs):
    raise ApiError(
        status_code=503,
        body="""
        <html><head>
        <meta http-equiv="content-type" content="text/html;charset=utf-8">
        <title>503 Server Error</title>
        </head>
        <body text=#000000 bgcolor=#ffffff>
        <h1>Error: Server Error</h1>
        <h2>The service you requested is not available at this time.<p>Service error -27.</h2>
        <h2></h2>
        </body></html>
        """,
    )


def raise_api_error_with_json(*args, **kwargs):
    raise ApiError(
        status_code=422,
        body=dict(detail="Please use numbers"),
    )


def raise_read_timeout_error(*args, **kwargs):
    sleep_seconds = 5
    print(f"raising ReadTimeout error after {sleep_seconds} seconds")
    time.sleep(sleep_seconds)
    raise httpx.ReadTimeout("Timed out")


def raise_http_error(*args, **kwargs):
    print("raising HTTP error")
    raise ApiError(status_code=503, body="HTTP error")


@pytest.mark.parametrize(
    "side_effect,side_effect_exception,should_retry",
    [
        (raise_api_error_with_text, ApiError, True),
        (raise_api_error_with_json, ApiError, False),
    ],
)
def test_retry_behavior(
    air_passengers_df, side_effect, side_effect_exception, should_retry
):
    check_retry_behavior(
        df=air_passengers_df,
        side_effect=side_effect,
        side_effect_exception=side_effect_exception,
        should_retry=should_retry,
    )


combs = [
    (2, 5, 30),
    (10, 1, 5),
]
side_effect_settings = [
    (raise_read_timeout_error, httpx.ReadTimeout),
    (raise_http_error, ApiError),
]


@pytest.mark.parametrize(
    "retry_settings,side_effect_settings", product(combs, side_effect_settings)
)
def test_retry_behavior_set2(air_passengers_df, retry_settings, side_effect_settings):
    max_retries, retry_interval, max_wait_time = retry_settings
    side_effect, side_effect_exception = side_effect_settings
    check_retry_behavior(
        df=air_passengers_df,
        side_effect=side_effect,
        side_effect_exception=side_effect_exception,
        max_retries=max_retries,
        retry_interval=retry_interval,
        max_wait_time=max_wait_time,
    )



import sys
from time import time

import fire
from dotenv import load_dotenv
from nixtla import NixtlaClient

from src.tools import ExperimentHandler

load_dotenv()


def evaluate_experiment(file: str):
    exp_handler = ExperimentHandler(file=file, method="timegpt")
    model_name = "TimeGPT"
    print(model_name)
    # timegpt does not need the full history to
    # make zero shot predictions
    Y_df, freq, pandas_freq, h, seasonality = exp_handler.read_data(
        max_insample_length=300
    )
    size_df = sys.getsizeof(Y_df) / (1024 * 1024)
    max_partition_size_mb = 20
    num_partitions = int(size_df / max_partition_size_mb) + 1
    timegpt = NixtlaClient(
        base_url="https://timegpt-endpoint.eastus.inference.ml.azure.com/",
        max_retries=1,
    )
    start = time()
    Y_hat_df = timegpt.cross_validation(
        df=Y_df,
        h=h,
        n_windows=1,
        freq=pandas_freq,
        num_partitions=num_partitions,
    )
    total_time = time() - start
    print(total_time)
    # evaluation
    eval_df, total_time_df = exp_handler.evaluate_model(
        Y_hat_df=Y_hat_df,
        model_name=model_name,
        total_time=total_time,
    )
    exp_handler.save_results(
        freq=freq,
        eval_df=eval_df,
        total_time_df=total_time_df,
    )


if __name__ == "__main__":
    fire.Fire(evaluate_experiment)



import os

import pandas as pd
from statsforecast import StatsForecast
from statsforecast.models import (
    _TS as StatsForecastModel,
    ADIDA as _ADIDA,
    AutoARIMA as _AutoARIMA,
    AutoCES as _AutoCES,
    AutoETS as _AutoETS,
    CrostonClassic as _CrostonClassic,
    DynamicOptimizedTheta as _DOTheta,
    HistoricAverage as _HistoricAverage,
    IMAPA as _IMAPA,
    SeasonalNaive as _SeasonalNaive,
    Theta as _Theta,
    ZeroModel as _ZeroModel,
)

from ..utils.forecaster import Forecaster, get_seasonality

os.environ["NIXTLA_ID_AS_COL"] = "true"


def run_statsforecast_model(
    model: StatsForecastModel,
    df: pd.DataFrame,
    h: int,
    freq: str,
) -> pd.DataFrame:
    sf = StatsForecast(
        models=[model],
        freq=freq,
        n_jobs=-1,
        fallback_model=_SeasonalNaive(
            season_length=get_seasonality(freq),
        ),
    )
    fcst_df = sf.forecast(df=df, h=h)
    return fcst_df


class ADIDA(Forecaster):
    def __init__(
        self,
        alias: str = "ADIDA",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        fcst_df = run_statsforecast_model(
            model=_ADIDA(alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class AutoARIMA(Forecaster):
    def __init__(
        self,
        alias: str = "AutoARIMA",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_AutoARIMA(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class AutoCES(Forecaster):
    def __init__(
        self,
        alias: str = "AutoCES",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_AutoCES(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class AutoETS(Forecaster):
    def __init__(
        self,
        alias: str = "AutoETS",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_AutoETS(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class CrostonClassic(Forecaster):
    def __init__(
        self,
        alias: str = "CrostonClassic",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        fcst_df = run_statsforecast_model(
            model=_CrostonClassic(alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class DOTheta(Forecaster):
    def __init__(
        self,
        alias: str = "DOTheta",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_DOTheta(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class HistoricAverage(Forecaster):
    def __init__(
        self,
        alias: str = "HistoricAverage",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        fcst_df = run_statsforecast_model(
            model=_HistoricAverage(alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class IMAPA(Forecaster):
    def __init__(
        self,
        alias: str = "IMAPA",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        fcst_df = run_statsforecast_model(
            model=_IMAPA(alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class SeasonalNaive(Forecaster):
    def __init__(
        self,
        alias: str = "SeasonalNaive",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_SeasonalNaive(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class Theta(Forecaster):
    def __init__(
        self,
        alias: str = "Theta",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        seasonality = get_seasonality(freq)
        fcst_df = run_statsforecast_model(
            model=_Theta(season_length=seasonality, alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df


class ZeroModel(Forecaster):
    def __init__(
        self,
        alias: str = "ZeroModel",
    ):
        self.alias = alias

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        fcst_df = run_statsforecast_model(
            model=_ZeroModel(alias=self.alias),
            df=df,
            h=h,
            freq=freq,
        )
        return fcst_df



import pytest

try:
    import dask.dataframe as dd
    from dask.distributed import Client

    @pytest.fixture(scope="module")
    def dask_client():
        with Client() as client:
            yield client

    @pytest.fixture(scope="module")
    def dask_df(distributed_series):
        return dd.from_pandas(distributed_series, npartitions=2)

    @pytest.fixture(scope="module")
    def dask_diff_cols_df(distributed_series, renamer):
        return dd.from_pandas(
            distributed_series.rename(columns=renamer),
            npartitions=2,
        )

    @pytest.fixture(scope="module")
    def dask_df_x(distributed_df_x):
        return dd.from_pandas(distributed_df_x, npartitions=2)

    @pytest.fixture(scope="module")
    def dask_future_ex_vars_df(distributed_future_ex_vars_df):
        return dd.from_pandas(distributed_future_ex_vars_df, npartitions=2)

    @pytest.fixture(scope="module")
    def dask_df_x_diff_cols(distributed_df_x, renamer):
        return dd.from_pandas(distributed_df_x.rename(columns=renamer), npartitions=2)

    @pytest.fixture(scope="module")
    def dask_future_ex_vars_df_diff_cols(distributed_future_ex_vars_df, renamer):
        return dd.from_pandas(
            distributed_future_ex_vars_df.rename(columns=renamer), npartitions=2
        )
except ImportError:
    # If Dask is not installed, we skip the fixtures
    pytest.skip(
        "Dask is not installed, skipping Dask fixtures", allow_module_level=True
    )



import os
from copy import deepcopy
from types import SimpleNamespace

import numpy as np
import pandas as pd
import pytest
import utilsforecast.processing as ufp
from dotenv import load_dotenv
from utilsforecast.data import generate_series
from utilsforecast.feature_engineering import fourier, time_features

from nixtla.nixtla_client import NixtlaClient, _maybe_add_date_features
from nixtla_tests.helpers.states import model_ids_object

load_dotenv(override=True)

pytest_plugins = [
    "nixtla_tests.fixtures.dask_fixtures",
    "nixtla_tests.fixtures.spark_fixtures",
    "nixtla_tests.fixtures.ray_fixtures",
]


# note that scope="session" will result in failed test
@pytest.fixture(scope="module")
def nixtla_test_client():
    client = NixtlaClient()
    yield client

    try:
        client.delete_finetuned_model(model_ids_object.model_id1)
    except:
        print("model_id1 not found, skipping deletion.")

    try:
        client.delete_finetuned_model(model_ids_object.model_id2)
    except:
        print("model_id2 not found, skipping deletion.")


@pytest.fixture(scope="class")
def custom_client():
    client = NixtlaClient(
        base_url=os.environ["NIXTLA_BASE_URL_CUSTOM"],
        api_key=os.environ["NIXTLA_API_KEY_CUSTOM"],
    )
    yield client

    try:
        client.delete_finetuned_model(model_ids_object.model_id1)
    except:
        print("model_id1 not found, skipping deletion.")

    try:
        client.delete_finetuned_model(model_ids_object.model_id2)
    except:
        print("model_id2 not found, skipping deletion.")


@pytest.fixture
def series_with_gaps():
    series = generate_series(2, min_length=100, freq="5min")
    with_gaps = series.sample(frac=0.5, random_state=0)
    return series, with_gaps


@pytest.fixture
def df_no_duplicates():
    return pd.DataFrame(
        {
            "unique_id": [1, 2, 3, 4],
            "ds": ["2020-01-01"] * 4,
            "y": [1, 2, 3, 4],
        }
    )


@pytest.fixture
def df_with_duplicates():
    return pd.DataFrame(
        {
            "unique_id": [1, 1, 1],
            "ds": ["2020-01-01", "2020-01-01", "2020-01-02"],
            "y": [1, 2, 3],
        }
    )


@pytest.fixture
def df_complete():
    return pd.DataFrame(
        {
            "unique_id": [1, 1, 1, 2, 2, 2],
            "ds": [
                "2020-01-01",
                "2020-01-02",
                "2020-01-03",
                "2020-01-01",
                "2020-01-02",
                "2020-01-03",
            ],
            "y": [1, 2, 3, 4, 5, 6],
        }
    )


@pytest.fixture
def df_missing():
    return pd.DataFrame(
        {
            "unique_id": [1, 1, 2, 2],
            "ds": ["2020-01-01", "2020-01-03", "2020-01-01", "2020-01-03"],
            "y": [1, 3, 4, 6],
        }
    )


@pytest.fixture
def df_no_cat():
    return pd.DataFrame(
        {
            "unique_id": [1, 2, 3],
            "ds": pd.date_range("2023-01-01", periods=3),
            "y": [1.0, 2.0, 3.0],
        }
    )


@pytest.fixture
def df_with_cat():
    return pd.DataFrame(
        {
            "unique_id": ["A", "B", "C"],
            "ds": pd.date_range("2023-01-01", periods=3),
            "y": [1.0, 2.0, 3.0],
            "cat_col": ["X", "Y", "Z"],
        }
    )


@pytest.fixture
def df_with_cat_dtype():
    return pd.DataFrame(
        {
            "unique_id": [1, 2, 3],
            "ds": pd.date_range("2023-01-01", periods=3),
            "y": [1.0, 2.0, 3.0],
            "cat_col": pd.Categorical(["X", "Y", "Z"]),
        }
    )


@pytest.fixture
def df_leading_zeros():
    return pd.DataFrame(
        {
            "unique_id": ["A", "A", "A", "B", "B", "C", "C", "C", "C", "D", "D", "D"],
            "ds": pd.date_range("2025-01-01", periods=12),
            "y": [0, 1, 2, 0, 1, 0, 0, 1, 2, 0, 0, 0],
        }
    )


@pytest.fixture
def df_negative_values():
    return pd.DataFrame(
        {
            "unique_id": ["A", "A", "A", "B", "B", "C", "C", "C"],
            "ds": pd.date_range("2025-01-01", periods=8),
            "y": [0, -1, 2, -1, -2, 0, 1, 2],
        }
    )


@pytest.fixture(scope="module")
def ts_data_set1():
    h = 5
    freq = "D"
    series = generate_series(n_series=10, freq=freq, equal_ends=True)
    train_end = series["ds"].max() - h * pd.offsets.Day()
    train_mask = series["ds"] <= train_end
    train = series[train_mask]
    valid = series[~train_mask]

    return SimpleNamespace(
        h=h,
        series=series,
        train=train,
        train_end=train_end,
        valid=valid,
        freq=freq,
    )


@pytest.fixture(scope="module")
def ts_anomaly_data(ts_data_set1):
    train_anomalies = ts_data_set1.train.copy()
    anomaly_date = ts_data_set1.train_end - 2 * pd.offsets.Day()
    train_anomalies.loc[train_anomalies["ds"] == anomaly_date, "y"] *= 2

    return SimpleNamespace(
        train_anomalies=train_anomalies,
        anomaly_date=anomaly_date,
    )


@pytest.fixture
def common_kwargs():
    return {"freq": "D", "id_col": "unique_id", "time_col": "ds"}


@pytest.fixture
def df_ok():
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id1", "id2", "id2", "id2"],
            "ds": [
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
            ],
            "y": [1, 2, 3, 4, 5, 6],
        }
    )


@pytest.fixture
def df_with_duplicates_set2():
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id1", "id2"],
            "ds": ["2023-01-01", "2023-01-01", "2023-01-02", "2023-01-02"],
            "y": [1, 2, 3, 4],
        }
    )


@pytest.fixture
def df_with_missing_dates():
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id2", "id2"],
            "ds": ["2023-01-01", "2023-01-03", "2023-01-01", "2023-01-03"],
            "y": [1, 3, 4, 6],
        }
    )


@pytest.fixture
def df_with_duplicates_and_missing_dates():
    # Global end on 2023-01-03 which is missing for id1
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id1", "id2", "id2"],
            "ds": [
                "2023-01-01",
                "2023-01-01",
                "2023-01-02",
                "2023-01-02",
                "2023-01-03",
            ],
            "y": [1, 2, 3, 4, 5],
        }
    )


@pytest.fixture
def df_with_cat_columns():
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id1", "id2", "id2", "id2"],
            "ds": [
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
            ],
            "y": [1, 2, 3, 4, 5, 6],
            "cat_col1": ["A", "B", "C", "D", "E", "F"],
            "cat_col2": pd.Categorical(["X", "Y", "Z", "X", "Y", "Z"]),
        }
    )


@pytest.fixture
def df_negative_vals():
    return pd.DataFrame(
        {
            "unique_id": ["id1", "id1", "id1", "id2", "id2", "id2"],
            "ds": [
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
            ],
            "y": [-1, 0, 1, 2, -3, -4],
        }
    )


@pytest.fixture
def df_leading_zeros_set2():
    return pd.DataFrame(
        {
            "unique_id": [
                "id1",
                "id1",
                "id1",
                "id2",
                "id2",
                "id2",
                "id3",
                "id3",
                "id3",
            ],
            "ds": [
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
                "2023-01-01",
                "2023-01-02",
                "2023-01-03",
            ],
            "y": [0, 1, 2, 0, 1, 2, 0, 0, 0],
        }
    )


@pytest.fixture
def cv_series_with_features():
    freq = "D"
    h = 5
    series = generate_series(2, freq=freq)
    series_with_features, _ = fourier(series, freq=freq, season_length=7, k=2)
    splits = ufp.backtest_splits(
        df=series_with_features,
        n_windows=1,
        h=h,
        id_col="unique_id",
        time_col="ds",
        freq=freq,
    )
    _, train, valid = next(splits)
    x_cols = train.columns.drop(["unique_id", "ds", "y"]).tolist()
    return series_with_features, train, valid, x_cols, h, freq


@pytest.fixture
def custom_business_hours():
    return pd.tseries.offsets.CustomBusinessHour(
        start="09:00",
        end="16:00",
        holidays=[
            "2022-12-25",  # Christmas
            "2022-01-01",  # New Year's Day
        ],
    )


@pytest.fixture
def business_hours_series(custom_business_hours):
    series = pd.DataFrame(
        {
            "unique_id": 1,
            "ds": pd.date_range(
                start="2000-01-03 09", freq=custom_business_hours, periods=200
            ),
            "y": np.arange(200) % 7,
        }
    )
    series = pd.concat([series.assign(unique_id=i) for i in range(10)]).reset_index(
        drop=True
    )
    return series


@pytest.fixture
def integer_freq_series():
    series = generate_series(5, freq="H", min_length=200)
    series["ds"] = series.groupby("unique_id", observed=True)["ds"].cumcount()
    return series


@pytest.fixture
def two_short_series():
    return generate_series(n_series=2, min_length=5, max_length=20)


@pytest.fixture
def two_short_series_with_time_features_train_future(two_short_series):
    train, future = time_features(
        two_short_series, freq="D", features=["year", "month"], h=5
    )
    return train, future


@pytest.fixture
def series_1MB_payload():
    series = generate_series(250, n_static_features=2)
    return series


@pytest.fixture(scope="module")
def air_passengers_df():
    return pd.read_csv(
        "https://raw.githubusercontent.com/Nixtla/transfer-learning-time-series/main/datasets/air_passengers.csv",
        parse_dates=["timestamp"],
    )


@pytest.fixture(scope="module")
def air_passengers_renamed_df(air_passengers_df):
    df_copy = deepcopy(air_passengers_df)
    df_copy.rename(columns={"timestamp": "ds", "value": "y"}, inplace=True)
    df_copy.insert(0, "unique_id", "AirPassengers")
    return df_copy


@pytest.fixture(scope="module")
def air_passengers_renamed_df_with_index(air_passengers_renamed_df):
    df_copy = deepcopy(air_passengers_renamed_df)
    df_ds_index = df_copy.set_index("ds")[["unique_id", "y"]]
    df_ds_index.index = pd.DatetimeIndex(df_ds_index.index)
    return df_ds_index


@pytest.fixture
def df_freq_generator():
    def _df_freq(n_series, min_length, max_length, freq):
        df_freq = generate_series(
            n_series,
            min_length=min_length if freq != "15T" else 1_200,
            max_length=max_length if freq != "15T" else 2_000,
        )
        return df_freq

    return _df_freq


@pytest.fixture(scope="module")
def date_features_result(air_passengers_renamed_df):
    date_features = ["year", "month"]
    df_date_features, future_df = _maybe_add_date_features(
        df=air_passengers_renamed_df,
        X_df=None,
        h=12,
        freq="MS",
        features=date_features,
        one_hot=False,
        id_col="unique_id",
        time_col="ds",
        target_col="y",
    )
    return df_date_features, future_df, date_features


HYPER_PARAMS_TEST = [
    # finetune steps is unstable due
    # to numerical reasons
    # dict(finetune_steps=2),
    dict(),
    dict(clean_ex_first=False),
    dict(date_features=["month"]),
    dict(level=[80, 90]),
    # dict(level=[80, 90], finetune_steps=2),
]


@pytest.fixture(scope="module")
def train_test_split(air_passengers_renamed_df):
    df_ = deepcopy(air_passengers_renamed_df)
    df_test = df_.groupby("unique_id").tail(12)
    df_train = df_.drop(df_test.index)
    return df_train, df_test


@pytest.fixture(scope="module")
def exog_data(air_passengers_renamed_df, train_test_split):
    df_ = deepcopy(air_passengers_renamed_df)
    df_ex_ = df_.copy()
    df_ex_["exogenous_var"] = df_ex_["y"] + np.random.normal(size=len(df_ex_))
    df_train, df_test = train_test_split
    x_df_test = df_test.drop(columns="y").merge(df_ex_.drop(columns="y"))
    return df_ex_, df_train, df_test, x_df_test


@pytest.fixture(scope="module")
def large_series():
    return generate_series(20_000, min_length=1_000, max_length=1_000, freq="min")


@pytest.fixture(scope="module")
def anomaly_online_df():
    detection_size = 5
    n_series = 2
    size = 100
    ds = pd.date_range(start="2023-01-01", periods=size, freq="W")
    x = np.arange(size)
    y = 10 * np.sin(0.1 * x) + 12
    y = np.tile(y, n_series)
    y[size - 5] = 30
    y[2 * size - 1] = 30
    df = pd.DataFrame(
        {
            "unique_id": np.repeat(np.arange(1, n_series + 1), size),
            "ds": np.tile(ds, n_series),
            "y": y,
        }
    )
    return df, n_series, detection_size


@pytest.fixture(scope="module")
def distributed_n_series():
    return 4


@pytest.fixture(scope="module")
def renamer():
    return {
        "unique_id": "id_col",
        "ds": "time_col",
        "y": "target_col",
    }


@pytest.fixture(scope="module")
def distributed_series(distributed_n_series):
    series = generate_series(distributed_n_series, min_length=100)
    series["unique_id"] = series["unique_id"].astype(str)
    return series


@pytest.fixture(scope="module")
def distributed_df_x():
    df_x = pd.read_csv(
        "https://raw.githubusercontent.com/Nixtla/transfer-learning-time-series/main/datasets/electricity-short-with-ex-vars.csv",
        parse_dates=["ds"],
    ).rename(columns=str.lower)
    return df_x


@pytest.fixture(scope="module")
def distributed_future_ex_vars_df():
    future_ex_vars_df = pd.read_csv(
        "https://raw.githubusercontent.com/Nixtla/transfer-learning-time-series/main/datasets/electricity-short-future-ex-vars.csv",
        parse_dates=["ds"],
    ).rename(columns=str.lower)
    return future_ex_vars_df



from pathlib import Path

import fire
import pandas as pd
from azure.ai.ml.entities import AmlCompute

from .automl_handler import AzureAutoML
from src.utils.data_handler import ExperimentDataset


class AzureAutoMLJobs:
    """
    This class stores and updates the Azure AutoML Experiments,
    to keep track of the pipeline jobs.
    We need this to later downlaod the forecasts.
    """

    file_name = "forecasting_jobs.csv"

    def __init__(self, dir: str = "./azure_automl_results"):
        self.dir = dir
        self.jobs_path = Path(self.dir) / self.file_name
        self.setup()

    def setup(self):
        self.jobs_path.parent.mkdir(parents=True, exist_ok=True)
        if not self.jobs_path.exists():
            pd.DataFrame(columns=["created_at", "experiment_name", "job_name"]).to_csv(
                self.jobs_path,
                index=False,
            )

    def get_jobs_df(self) -> pd.DataFrame:
        return pd.read_csv(self.jobs_path)

    def save_job(self, job_name: str, experiment_name: str):
        jobs_df = self.get_jobs_df()
        new_row = pd.DataFrame(
            {
                "created_at": [pd.Timestamp.now()],
                "experiment_name": [experiment_name],
                "job_name": [job_name],
            }
        )
        jobs_df = pd.concat([jobs_df, new_row])
        jobs_df.to_csv(self.jobs_path, index=False)


def start_forecasting_job(
    dataset_path: str,
    begin_create_or_update_aml_compute: bool = False,
):
    experiment_name = dataset_path.split("/")[-1].split(".")[0]
    dataset = ExperimentDataset.from_parquet(parquet_path=dataset_path)
    azure_automl = AzureAutoML.from_environment()
    azure_automl_jobs = AzureAutoMLJobs()

    aml_compute = AmlCompute(
        name="azure-automl-fcst-cluster-nixtla",
        min_instances=11,
        max_instances=11,
        size="STANDARD_DS5_V2",
    )

    job_name = azure_automl.forecast(
        df=dataset.Y_df_train,
        df_test=dataset.Y_df_test,
        aml_compute=aml_compute,
        h=dataset.horizon,
        freq=dataset.pandas_frequency,
        n_cross_validations=2,
        experiment_name=experiment_name,
        begin_create_or_update_aml_compute=begin_create_or_update_aml_compute,
        max_nodes=11,
        max_concurrency_per_node=8,
    )

    azure_automl_jobs.save_job(job_name, experiment_name)


if __name__ == "__main__":
    fire.Fire(start_forecasting_job)



from functools import partial
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
from gluonts.dataset import Dataset
from gluonts.dataset.repository.datasets import (
    get_dataset,
    dataset_names as gluonts_datasets,
)
from gluonts.time_feature.seasonality import get_seasonality
from utilsforecast.evaluation import evaluate
from utilsforecast.losses import mase, smape


def quantile_loss(
    df: pd.DataFrame,
    models: list,
    q: float = 0.5,
    id_col: str = "unique_id",
    target_col: str = "y",
) -> pd.DataFrame:
    delta_y = df[models].sub(df[target_col], axis=0)
    res = (
        np.maximum(q * delta_y, (q - 1) * delta_y)
        .groupby(df[id_col], observed=True)
        .mean()
    )
    res.index.name = id_col
    res = res.reset_index()
    return res


class ExperimentHandler:
    def __init__(
        self,
        dataset: str,
        quantiles: List[float] = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
        results_dir: str = "./results",
        models_dir: str = "./models",
    ):
        if dataset not in gluonts_datasets:
            raise Exception(
                f"dataset {dataset} not found in gluonts "
                f"available datasets: {', '.join(gluonts_datasets)}"
            )
        self.dataset = dataset
        self.quantiles = quantiles
        self.level = self._transform_quantiles_to_levels(quantiles)
        self.results_dir = results_dir
        self.models_dir = models_dir
        # defining datasets
        self._maybe_download_m3_file(self.dataset)
        gluonts_dataset = get_dataset(self.dataset)
        self.horizon = gluonts_dataset.metadata.prediction_length
        if self.horizon is None:
            raise Exception(
                f"horizon not found for dataset {self.dataset} "
                "experiment cannot be run"
            )
        self.freq = gluonts_dataset.metadata.freq
        self.seasonality = get_seasonality(self.freq)
        self.gluonts_train_dataset = gluonts_dataset.train
        self.gluonts_test_dataset = gluonts_dataset.test
        self._create_dir_if_not_exists(self.results_dir)

    @staticmethod
    def _maybe_download_m3_file(dataset: str):
        if dataset[:2] == "m3":
            m3_file = Path.home() / ".gluonts" / "datasets" / "M3C.xls"
            if not m3_file.exists():
                from datasetsforecast.m3 import M3
                from datasetsforecast.utils import download_file

                download_file(m3_file.parent, M3.source_url)

    @staticmethod
    def _transform_quantiles_to_levels(quantiles: List[float]) -> List[int]:
        level = [
            int(100 - 200 * q) for q in quantiles if q < 0.5
        ]  # in this case mean=mediain
        level = sorted(list(set(level)))
        return level

    @staticmethod
    def _create_dir_if_not_exists(directory: str):
        Path(directory).mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _transform_gluonts_instance_to_df(
        ts: dict,
        last_n: int | None = None,
    ) -> pd.DataFrame:
        start_period = ts["start"]
        start_ds, freq = start_period.to_timestamp(), start_period.freq
        target = ts["target"]
        ds = pd.date_range(start=start_ds, freq=freq, periods=len(target))
        if last_n is not None:
            target = target[-last_n:]
            ds = ds[-last_n:]
        ts_df = pd.DataFrame({"unique_id": ts["item_id"], "ds": ds, "y": target})
        return ts_df

    @staticmethod
    def _transform_gluonts_dataset_to_df(
        gluonts_dataset: Dataset,
        last_n: int | None = None,
    ) -> pd.DataFrame:
        df = pd.concat(
            [
                ExperimentHandler._transform_gluonts_instance_to_df(ts, last_n=last_n)
                for ts in gluonts_dataset
            ]
        )
        df = df.reset_index(drop=True)
        return df

    @property
    def train_df(self) -> pd.DataFrame:
        train_df = self._transform_gluonts_dataset_to_df(self.gluonts_train_dataset)
        return train_df

    @property
    def test_df(self) -> pd.DataFrame:
        test_df = self._transform_gluonts_dataset_to_df(
            self.gluonts_test_dataset,
            last_n=self.horizon,
        )
        return test_df

    def save_dataframe(self, df: pd.DataFrame, file_name: str):
        df.to_csv(f"{self.results_dir}/{file_name}", index=False)

    def save_results(self, fcst_df: pd.DataFrame, total_time: float, model_name: str):
        self.save_dataframe(
            fcst_df,
            f"{model_name}-{self.dataset}-fcst.csv",
        )
        time_df = pd.DataFrame({"time": [total_time], "model": model_name})
        self.save_dataframe(
            time_df,
            f"{model_name}-{self.dataset}-time.csv",
        )

    def fcst_from_level_to_quantiles(
        self,
        fcst_df: pd.DataFrame,
        model_name: str,
    ) -> pd.DataFrame:
        fcst_df = fcst_df.copy()
        cols = ["unique_id", "ds", model_name]
        for q in self.quantiles:
            if q == 0.5:
                col = f"{model_name}"
            else:
                lv = int(100 - 200 * q)
                hi_or_lo = "lo" if lv > 0 else "hi"
                lv = abs(lv)
                col = f"{model_name}-{hi_or_lo}-{lv}"
            q_col = f"{model_name}-q-{q}"
            fcst_df[q_col] = fcst_df[col].values
            cols.append(q_col)
        return fcst_df[cols]

    def evaluate_models(self, models: List[str]) -> pd.DataFrame:
        test_df = self.test_df
        train_df = self.train_df
        fcsts_df = []
        times_df = []
        for model in models:
            fcst_method_df = pd.read_csv(
                f"{self.results_dir}/{model}-{self.dataset}-fcst.csv"
            ).set_index(["unique_id", "ds"])
            fcsts_df.append(fcst_method_df)
            time_method_df = pd.read_csv(
                f"{self.results_dir}/{model}-{self.dataset}-time.csv"
            )
            times_df.append(time_method_df)
        fcsts_df = pd.concat(fcsts_df, axis=1).reset_index()
        fcsts_df["ds"] = pd.to_datetime(fcsts_df["ds"])
        times_df = pd.concat(times_df)
        test_df = test_df.merge(fcsts_df, how="left")
        assert test_df.isna().sum().sum() == 0, "merge contains nas"
        # point evaluation
        point_fcsts_cols = ["unique_id", "ds", "y"] + models
        test_df["unique_id"] = test_df["unique_id"].astype(str)
        train_df["unique_id"] = train_df["unique_id"].astype(str)
        mase_seas = partial(mase, seasonality=self.seasonality)
        eval_df = evaluate(
            test_df[point_fcsts_cols],
            train_df=train_df,
            metrics=[smape, mase_seas],
        )
        # probabilistic evaluation
        eval_prob_df = []
        for q in self.quantiles:
            prob_cols = [f"{model}-q-{q}" for model in models]
            eval_q_df = quantile_loss(test_df, models=prob_cols, q=q)
            eval_q_df[prob_cols] = eval_q_df[prob_cols] * self.horizon
            eval_q_df = eval_q_df.rename(columns=dict(zip(prob_cols, models)))
            eval_q_df["metric"] = f"quantile-loss-{q}"
            eval_prob_df.append(eval_q_df)
        eval_prob_df = pd.concat(eval_prob_df)
        eval_prob_df = eval_prob_df.groupby("metric").sum().reset_index()
        total_y = test_df["y"].sum()
        eval_prob_df[models] = eval_prob_df[models] / total_y
        eval_prob_df["metric"] = "scaled_crps"
        eval_df = pd.concat([eval_df, eval_prob_df]).reset_index(drop=True)
        eval_df = eval_df.groupby("metric").mean(numeric_only=True).reset_index()
        eval_df = eval_df.melt(id_vars="metric", value_name="value", var_name="model")
        times_df.insert(0, "metric", "time")
        times_df = times_df.rename(columns={"time": "value"})
        eval_df = pd.concat([eval_df, times_df])
        eval_df.insert(0, "dataset", self.dataset)
        eval_df = eval_df.sort_values(["dataset", "metric", "model"])
        eval_df = eval_df.reset_index(drop=True)
        return eval_df



import pandas as pd

from src.main import vn1_competition_evaluation, get_competition_forecasts


def test_vn1_competition_evaluation():
    forecasts = get_competition_forecasts()
    eval_df = vn1_competition_evaluation(forecasts)
    assert len(eval_df) == 5
    pd.testing.assert_series_equal(
        eval_df["score"], 
        pd.Series([0.4637, 0.4657, 0.4758, 0.4774, 0.4808]),
        check_names=False,
    )



import os
from pathlib import Path
from time import time

import fire
from statsforecast import StatsForecast
from statsforecast.models import SeasonalNaive

from src.utils.data_handler import ExperimentDataset, ForecastDataset


def sn_forecast(dataset_path: str, results_dir: str = "./results"):
    os.environ["NIXTLA_ID_AS_COL"] = "true"
    dataset = ExperimentDataset.from_parquet(parquet_path=dataset_path)
    sf = StatsForecast(
        models=[SeasonalNaive(season_length=dataset.seasonality)],
        freq=dataset.pandas_frequency,
    )
    start = time()
    forecast_df = sf.forecast(
        df=dataset.Y_df_train,
        h=dataset.horizon,
    )
    end = time()
    total_time = end - start
    forecast_dataset = ForecastDataset(forecast_df=forecast_df, total_time=total_time)
    experiment_name = dataset_path.split("/")[-1].split(".")[0]
    results_path = Path(results_dir) / "statsforecast_sn" / experiment_name
    forecast_dataset.save_to_dir(results_path)


if __name__ == "__main__":
    fire.Fire(sn_forecast)



from time import time
from typing import Iterable, List, Tuple

import fire
import pandas as pd
import torch
from gluonts.dataset import Dataset
from gluonts.model.forecast import Forecast
from gluonts.torch.model.predictor import PyTorchPredictor
from huggingface_hub import hf_hub_download
from tqdm import tqdm
from uni2ts.model.moirai import MoiraiForecast

from src.utils import ExperimentHandler


def get_morai_predictor(
    model_size: str,
    prediction_length: int,
    target_dim: int,
    batch_size: int,
) -> PyTorchPredictor:
    model = MoiraiForecast.load_from_checkpoint(
        checkpoint_path=hf_hub_download(
            repo_id=f"Salesforce/moirai-1.0-R-{model_size}",
            filename="model.ckpt",
        ),
        prediction_length=prediction_length,
        context_length=200,
        patch_size="auto",
        num_samples=100,
        target_dim=target_dim,
        feat_dynamic_real_dim=0,
        past_feat_dynamic_real_dim=0,
        map_location="cuda:0" if torch.cuda.is_available() else "cpu",
    )

    predictor = model.create_predictor(batch_size)

    return predictor


def gluonts_instance_fcst_to_df(
    fcst: Forecast,
    quantiles: List[float],
    model_name: str,
) -> pd.DataFrame:
    point_forecast = fcst.mean
    h = len(point_forecast)
    dates = pd.date_range(
        fcst.start_date.to_timestamp(),
        freq=fcst.freq,
        periods=h,
    )
    fcst_df = pd.DataFrame(
        {
            "ds": dates,
            "unique_id": fcst.item_id,
            model_name: point_forecast,
        }
    )
    for q in quantiles:
        fcst_df[f"{model_name}-q-{q}"] = fcst.quantile(q)
    return fcst_df


def gluonts_fcsts_to_df(
    fcsts: Iterable[Forecast],
    quantiles: List[float],
    model_name: str,
) -> pd.DataFrame:
    df = []
    for fcst in tqdm(fcsts):
        fcst_df = gluonts_instance_fcst_to_df(fcst, quantiles, model_name)
        df.append(fcst_df)
    return pd.concat(df).reset_index(drop=True)


def run_moirai(
    gluonts_dataset: Dataset,
    model_size: str,
    horizon: int,
    target_dim: int,
    batch_size: int,
    quantiles: List[float],
) -> Tuple[pd.DataFrame, float, str]:
    init_time = time()
    predictor = get_morai_predictor(model_size, horizon, target_dim, batch_size)
    fcsts = predictor.predict(gluonts_dataset)
    model_name = "SalesforceMoirai"
    fcsts_df = gluonts_fcsts_to_df(
        fcsts,
        quantiles=quantiles,
        model_name=model_name,
    )
    total_time = time() - init_time
    return fcsts_df, total_time, model_name


def main(dataset: str):
    exp = ExperimentHandler(dataset)
    fcst_df, total_time, model_name = run_moirai(
        gluonts_dataset=exp.gluonts_train_dataset,
        model_size="large",
        horizon=exp.horizon,
        target_dim=1,
        batch_size=32,
        quantiles=exp.quantiles,
    )
    exp.save_results(fcst_df, total_time, model_name)


if __name__ == "__main__":
    fire.Fire(main)



"""
this module takes Nixtla's benchmarking data 
and filters it to prevent azureml from crashing
in the following cases:
- too short series, see https://learn.microsoft.com/en-us/azure/machine-learning/concept-automl-forecasting-methods?view=azureml-api-2#data-length-requirements
"""
import logging
from pathlib import Path

import fire
import numpy as np
import pandas as pd

main_logger = logging.getLogger(__name__)
main_logger.setLevel(logging.INFO)


def get_min_size_per_series(dataset_path: str) -> int:
    if "Daily" in dataset_path or "Hourly" in dataset_path:
        return 1_000
    elif "Monthly" in dataset_path:
        return 10 * 12
    else:
        return 1_000 // 7


def filter_and_clean_dataset(
    dataset_path: str,
    max_series: int = 1_000,
    random_seed: int = 420,
):
    main_logger.info(f"Processing dataset {dataset_path}")
    df = pd.read_parquet(dataset_path)
    df = df.drop_duplicates(["unique_id", "ds"])  # type: ignore
    df = df.sort_values(["unique_id", "ds"])
    min_size_per_series = get_min_size_per_series(dataset_path)
    df = (
        df.groupby("unique_id")
        .filter(lambda x: len(x) >= min_size_per_series)
        .reset_index(drop=True)
    )
    uids = df["unique_id"].unique()  # type: ignore
    if len(uids) > max_series:
        np.random.seed(random_seed)
        uids = np.random.choice(uids, max_series, replace=False)  # type: ignore
        df = df.query("unique_id in @uids")  # type: ignore
        main_logger.info(f"Filtering out {len(uids) - max_series} series")
    n_series = len(df["unique_id"].unique())  # type: ignore
    main_logger.info(f"Number of series: {n_series}")
    if n_series == 0:
        raise ValueError("No series left after filtering")
    # finally we clean some strange dates
    mask = df["ds"].str.endswith(":01")  # type: ignore
    df.loc[mask, "ds"] = df.loc[mask, "ds"].str[:-3] + ":00"
    # save the dataset
    dataset_path = Path(dataset_path)  # type: ignore
    filtered_dataset_path = dataset_path.parent / "filtered_datasets" / dataset_path.name  # type: ignore
    filtered_dataset_path.parent.mkdir(exist_ok=True, parents=True)
    df.to_parquet(filtered_dataset_path)
    main_logger.info(f"Filtered dataset saved to {filtered_dataset_path}")


if __name__ == "__main__":
    fire.Fire(filter_and_clean_dataset)



import logging


def setup_logger(logger_name, log_file=None):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s,%(levelname)s,%(module)s,%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger



def test_detect_anomalies_online_univariate(nixtla_test_client, anomaly_online_df):
    df, n_series, detection_size = anomaly_online_df
    anomaly_df = nixtla_test_client.detect_anomalies_online(
        df,
        h=20,
        detection_size=detection_size,
        threshold_method="univariate",
        freq="W-SUN",
        level=99,
    )
    assert len(anomaly_df) == n_series * detection_size
    assert (
        len(anomaly_df.columns) == 8
    )  # [unique_id, ds, TimeGPT, y, anomaly, anomaly_score, hi, lo]
    assert anomaly_df["anomaly"].sum() == 2
    assert anomaly_df["anomaly"].iloc[0] and anomaly_df["anomaly"].iloc[-1]


def test_detect_anomalies_online_multivariate(nixtla_test_client, anomaly_online_df):
    df, n_series, detection_size = anomaly_online_df
    multi_anomaly_df = nixtla_test_client.detect_anomalies_online(
        df,
        h=20,
        detection_size=detection_size,
        threshold_method="multivariate",
        freq="W-SUN",
        level=99,
    )
    assert len(multi_anomaly_df) == n_series * detection_size
    assert (
        len(multi_anomaly_df.columns) == 7
    )  # [unique_id, ds, TimeGPT, y, anomaly, anomaly_score, accumulated_anomaly_score]
    assert multi_anomaly_df["anomaly"].sum() == 4
    assert (
        multi_anomaly_df["anomaly"].iloc[0]
        and multi_anomaly_df["anomaly"].iloc[4]
        and multi_anomaly_df["anomaly"].iloc[5]
        and multi_anomaly_df["anomaly"].iloc[9]
    )



from functools import partial
from typing import List

import pandas as pd
import pytest
from utilsforecast.data import generate_series
from utilsforecast.evaluation import evaluate
from utilsforecast.processing import (
    backtest_splits,
    drop_index_if_pandas,
    join,
    maybe_compute_sort_indices,
    take_rows,
    vertical_concat,
)

from xiuhmolpilli.models.utils.forecaster import maybe_convert_col_to_datetime
from xiuhmolpilli.utils.experiment_handler import (
    generate_train_cv_splits,
    mase,
    ExperimentDataset,
)
from .utils import models


def generate_train_cv_splits_from_backtest_splits(
    df: pd.DataFrame,
    n_windows: int,
    h: int,
    freq: str,
    step_size: int = 1,
):
    df = maybe_convert_col_to_datetime(df, "ds")
    # mlforecast cv code
    results = []
    sort_idxs = maybe_compute_sort_indices(df, "unique_id", "ds")
    if sort_idxs is not None:
        df = take_rows(df, sort_idxs)
    splits = backtest_splits(
        df,
        n_windows=n_windows,
        h=h,
        id_col="unique_id",
        time_col="ds",
        freq=pd.tseries.frequencies.to_offset(freq),
        step_size=h if step_size is None else step_size,
    )
    for _, (cutoffs, train, _) in enumerate(splits):
        train_cv = join(train, cutoffs, on="unique_id")
        results.append(train_cv)
    out = vertical_concat(results)
    out = drop_index_if_pandas(out)
    return out


def generate_exp_dataset(
    n_series,
    freq,
    return_df: bool = False,
) -> ExperimentDataset | pd.DataFrame:
    df = generate_series(n_series, freq=freq, min_length=12)
    df["unique_id"] = df["unique_id"].astype(str)
    df["frequency"] = "frequency"
    df["pandas_frequency"] = freq
    df["seasonality"] = 7
    df["horizon"] = 2
    if return_df:
        return df
    return ExperimentDataset.from_df(df)


def evaluate_cv_from_scratch(
    df: pd.DataFrame,
    fcst_df: pd.DataFrame,
    models: List[str],
    seasonality: int,
) -> pd.DataFrame:
    partial_mase = partial(mase, seasonality=seasonality)
    uids = df["unique_id"].unique()
    results = []
    for uid in uids:
        df_ = df.query("unique_id == @uid")
        fcst_df_ = fcst_df.query("unique_id == @uid")
        cutoffs = fcst_df_["cutoff"].unique()
        for cutoff in cutoffs:
            df__ = df_.query("ds <= @cutoff")
            fcst_df__ = fcst_df_.query("cutoff == @cutoff")
            eval_df = evaluate(
                df=fcst_df__,
                train_df=df__,
                metrics=[partial_mase],
                models=models,
            )
            eval_df["cutoff"] = cutoff
            results.append(eval_df)
    out = pd.concat(results)
    out = out[["unique_id", "cutoff", "metric"] + models]
    return out


def sort_df(df: pd.DataFrame, cols: List[str]) -> pd.DataFrame:
    return df.sort_values(cols).reset_index(drop=True)


@pytest.mark.parametrize(
    "freq,n_windows,h,step_size",
    [
        ("H", 3, 2, 1),
        ("H", 1, 12, None),
        ("MS", 3, 2, 2),
    ],
)
def test_generate_train_cv_splits(freq, n_windows, h, step_size):
    df = generate_series(n_series=5, freq=freq)
    df["unique_id"] = df["unique_id"].astype(int)
    df_cv = generate_train_cv_splits_from_backtest_splits(
        df=df,
        n_windows=n_windows,
        step_size=step_size,
        h=h,
        freq=freq,
    )
    cutoffs = df_cv[["unique_id", "cutoff"]].drop_duplicates()
    train_cv_splits = generate_train_cv_splits(
        df=df,
        cutoffs=cutoffs,
    )
    p_sort_df = partial(sort_df, cols=["unique_id", "cutoff", "ds"])
    pd.testing.assert_frame_equal(
        p_sort_df(df_cv),
        p_sort_df(train_cv_splits),
    )


@pytest.mark.parametrize("model", models)
def test_eval(model):
    freq = "H"
    exp_dataset = generate_exp_dataset(n_series=5, freq=freq)
    fcst_df = model.cross_validation(
        exp_dataset.df,
        h=exp_dataset.horizon,
        freq=exp_dataset.pandas_frequency,
    )
    eval_df = exp_dataset.evaluate_forecast_df(
        forecast_df=fcst_df,
        models=[model.alias],
    )
    eval_df_from_scratch = evaluate_cv_from_scratch(
        df=exp_dataset.df,
        fcst_df=fcst_df,
        models=[model.alias],
        seasonality=exp_dataset.seasonality,
    )
    p_sort_df = partial(sort_df, cols=["unique_id", "cutoff", "metric"])
    pd.testing.assert_frame_equal(
        p_sort_df(eval_df),
        p_sort_df(eval_df_from_scratch),
    )



import os

import pandas as pd
from neuralforecast import NeuralForecast
from neuralforecast.auto import (
    AutoNHITS as _AutoNHITS,
    AutoTFT as _AutoTFT,
)
from neuralforecast.common._base_model import BaseModel as NeuralForecastModel
from ray import tune

from ..utils.forecaster import Forecaster

os.environ["NIXTLA_ID_AS_COL"] = "true"


def run_neuralforecast_model(
    model: NeuralForecastModel,
    df: pd.DataFrame,
    freq: str,
) -> pd.DataFrame:
    nf = NeuralForecast(
        models=[model],
        freq=freq,
    )
    nf.fit(df=df)
    fcst_df = nf.predict()
    return fcst_df


class AutoNHITS(Forecaster):
    def __init__(
        self,
        alias: str = "AutoNHITS",
        num_samples: int = 10,
        backend: str = "optuna",
    ):
        self.alias = alias
        self.num_samples = num_samples
        self.backend = backend

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        config = _AutoNHITS.get_default_config(h=h, backend="ray")
        config["scaler_type"] = tune.choice(["robust"])

        if self.backend == "optuna":
            config = _AutoNHITS._ray_config_to_optuna(config)
        fcst_df = run_neuralforecast_model(
            model=_AutoNHITS(
                h=h,
                alias=self.alias,
                num_samples=self.num_samples,
                backend=self.backend,
                config=config,
            ),
            df=df,
            freq=freq,
        )
        return fcst_df


class AutoTFT(Forecaster):
    def __init__(
        self,
        alias: str = "AutoTFT",
        num_samples: int = 10,
        backend: str = "optuna",
    ):
        self.alias = alias
        self.num_samples = num_samples
        self.backend = backend

    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        config = _AutoTFT.get_default_config(h=h, backend="ray")
        config["scaler_type"] = tune.choice(["robust"])
        if self.backend == "optuna":
            config = _AutoTFT._ray_config_to_optuna(config)
        fcst_df = run_neuralforecast_model(
            model=_AutoTFT(
                h=h,
                alias=self.alias,
                num_samples=self.num_samples,
                backend=self.backend,
                config=config,
            ),
            df=df,
            freq=freq,
        )
        return fcst_df



from .chronos import Chronos
from .lagllama import LagLlama
from .moirai import Moirai
from .timegpt import TimeGPT
from .timesfm import TimesFM

__all__ = [
    "Chronos",
    "LagLlama",
    "Moirai",
    "TimeGPT",
    "TimesFM",
]



from pathlib import Path
from time import perf_counter
from typing import List

import pandas as pd
from rich.console import Console
from rich.table import Table

from .models.utils.forecaster import Forecaster
from .utils.experiment_handler import ExperimentDataset, ForecastDataset
from .utils.logger_config import setup_logger


main_logger = setup_logger(__name__)


def print_df_rich(df: pd.DataFrame):
    console = Console()
    table = Table()
    for col in df.select_dtypes(include=["float"]).columns:
        df[col] = df[col].apply(lambda x: f"{x:.3f}")
    for col in df.columns:
        table.add_column(col)
    for row in df.itertuples(index=False):
        table.add_row(*row)
    console.print(table)


def time_to_df(total_time: float, model_name: str) -> pd.DataFrame:
    return pd.DataFrame({"metric": ["time"], model_name: [total_time]})


class FoundationalTimeSeriesArena:
    def __init__(
        self,
        models: List[Forecaster],
        parquet_data_paths: List[str],
        results_dir: str = "./nixtla-foundational-time-series/results/",
    ):
        self.models = models
        self.parquet_data_paths: List[Path] = [
            Path(path) for path in parquet_data_paths
        ]
        self.results_dir = Path(results_dir)
        self.evaluation_path = self.results_dir / "evaluation.csv"

    def get_model_results_path(self, model_alias: str):
        return Path(self.results_dir) / model_alias

    def compete(self, overwrite: bool = False):
        complete_eval = []
        for parquet_path in self.parquet_data_paths:
            main_logger.info(f"Running on {parquet_path}")
            dataset = ExperimentDataset.from_parquet(parquet_path=parquet_path)
            dataset_results_path = self.results_dir / parquet_path.stem
            models_fcsts = None
            models_times = None
            for model in self.models:
                main_logger.info(f"Evaluating {model.alias}")
                model_results_path = dataset_results_path / model.alias
                is_forecast_ready = ForecastDataset.is_forecast_ready(
                    model_results_path
                )
                if not is_forecast_ready or overwrite:
                    main_logger.info(f"Forecasting {model.alias}")
                    start = perf_counter()
                    forecast_df = model.cross_validation(
                        df=dataset.df,
                        h=dataset.horizon,
                        freq=dataset.pandas_frequency,
                    )
                    total_time = perf_counter() - start
                    fcst_dataset = ForecastDataset(
                        forecast_df=forecast_df,
                        time_df=time_to_df(total_time, model.alias),
                    )
                    fcst_dataset.save_to_dir(dir=model_results_path)
                else:
                    main_logger.info(f"Loading {model.alias} forecast")
                    fcst_dataset = ForecastDataset.from_dir(model_results_path)
                if models_fcsts is None:
                    models_fcsts = fcst_dataset.forecast_df
                    models_times = fcst_dataset.time_df
                else:
                    models_fcsts = models_fcsts.merge(
                        fcst_dataset.forecast_df.drop(columns="y"),
                        how="left",
                        on=["unique_id", "cutoff", "ds"],
                    )
                    models_times = models_times.merge(
                        fcst_dataset.time_df,
                        how="left",
                        on=["metric"],
                    )
            main_logger.info("Evaluating forecasts")
            eval_df = dataset.evaluate_forecast_df(
                forecast_df=models_fcsts,
                models=[model.alias for model in self.models],
            )
            eval_df = eval_df.groupby(["metric"], as_index=False).mean(
                numeric_only=True
            )
            eval_df = pd.concat([models_times, eval_df])
            eval_df.insert(0, "dataset", parquet_path.stem)
            complete_eval.append(eval_df)
        complete_eval = pd.concat(complete_eval)
        complete_eval.to_csv(
            self.evaluation_path,
            index=False,
        )
        print_df_rich(complete_eval)


if __name__ == "__main__":
    import fire

    from .models.benchmarks import (
        ADIDA,
        AutoARIMA,
        AutoCES,
        AutoETS,
        AutoLGBM,
        AutoNHITS,
        AutoTFT,
        CrostonClassic,
        DOTheta,
        HistoricAverage,
        NixtlaProphet,
        IMAPA,
        SeasonalNaive,
        Theta,
        ZeroModel,
    )
    from .models.foundational import (
        Chronos,
        LagLlama,
        Moirai,
        TimeGPT,
        TimesFM,
    )

    frequencies = ["Hourly", "Daily", "Weekly", "Monthly"]
    files = [
        f"./nixtla-foundational-time-series/data/{freq}.parquet" for freq in frequencies
    ]
    arena = FoundationalTimeSeriesArena(
        models=[
            # naive
            SeasonalNaive(),
            HistoricAverage(),
            ZeroModel(),
            # statistical
            AutoARIMA(),
            NixtlaProphet(),
            AutoCES(),
            AutoETS(),
            Theta(),
            DOTheta(),
            ADIDA(),
            IMAPA(),
            CrostonClassic(),
            # ml
            AutoLGBM(),
            # neural
            AutoTFT(),
            AutoNHITS(),
            # foundational models
            Chronos(),
            LagLlama(),
            Moirai(),
            TimeGPT(),
            TimeGPT(model="timegpt-1-long-horizon", alias="TimeGPTLongHorizon"),
            TimesFM(),
        ],
        parquet_data_paths=files,
    )
    fire.Fire(arena.compete)



from time import time

import fire
import pandas as pd
from statsforecast import StatsForecast
from statsforecast.models import SeasonalNaive, ZeroModel

from src.tools import ExperimentHandler


def evaluate_experiment(file: str):
    exp_handler = ExperimentHandler(file=file, method="statsforecast")
    Y_df, freq, pandas_freq, h, seasonality = exp_handler.read_data()
    models = [
        SeasonalNaive(season_length=seasonality),
        ZeroModel(),
    ]
    # even though statsforecast can handle multiple models, we only use one
    # at a time to calculate time for each
    eval_df = []
    total_time_df = []
    for model in models:
        model_name = repr(model)
        print(model_name)
        sf = StatsForecast(
            models=[model],
            freq=pandas_freq,
            n_jobs=-1,
        )
        start = time()
        Y_hat_df_model = sf.cross_validation(
            df=Y_df,
            h=h,
            n_windows=1,
        ).reset_index()
        total_time = time() - start
        print(total_time)
        # evaluation
        eval_df_model, total_time_df_model = exp_handler.evaluate_model(
            Y_hat_df=Y_hat_df_model,
            model_name=model_name,
            total_time=total_time,
        )
        eval_df.append(eval_df_model.set_index(["metric", "unique_id"]))
        total_time_df.append(total_time_df_model)
    eval_df = pd.concat(eval_df, axis=1).reset_index()
    total_time_df = pd.concat(total_time_df)
    exp_handler.save_results(
        freq=freq,
        eval_df=eval_df,
        total_time_df=total_time_df,
        df=Y_df,
    )


if __name__ == "__main__":
    fire.Fire(evaluate_experiment)



from time import time
from typing import Iterable, List, Tuple

import fire
import pandas as pd
import torch
from gluonts.dataset import Dataset
from gluonts.model.forecast import Forecast
from gluonts.torch.model.predictor import PyTorchPredictor
from tqdm import tqdm

from lag_llama.gluon.estimator import LagLlamaEstimator
from src.utils import ExperimentHandler


def get_lag_llama_predictor(
    prediction_length: int, models_dir: str
) -> PyTorchPredictor:
    model_path = f"{models_dir}/lag-llama.ckpt"
    map_location = torch.device("cuda:0") if torch.cuda.is_available() else "cpu"
    if map_location == "cpu":
        raise ValueError("cpu is not supported in lagllama (there is a bug)")
    ckpt = torch.load(model_path, map_location=map_location)
    estimator_args = ckpt["hyper_parameters"]["model_kwargs"]
    # this context length is reported in the paper
    context_length = 32
    estimator = LagLlamaEstimator(
        ckpt_path=model_path,
        prediction_length=prediction_length,
        context_length=context_length,
        # estimator args
        input_size=estimator_args["input_size"],
        n_layer=estimator_args["n_layer"],
        n_embd_per_head=estimator_args["n_embd_per_head"],
        n_head=estimator_args["n_head"],
        scaling=estimator_args["scaling"],
        time_feat=estimator_args["time_feat"],
    )
    lightning_module = estimator.create_lightning_module()
    transformation = estimator.create_transformation()
    predictor = estimator.create_predictor(transformation, lightning_module)
    return predictor


def gluonts_instance_fcst_to_df(
    fcst: Forecast,
    quantiles: List[float],
    model_name: str,
) -> pd.DataFrame:
    point_forecast = fcst.mean
    h = len(point_forecast)
    dates = pd.date_range(
        fcst.start_date.to_timestamp(),
        freq=fcst.freq,
        periods=h,
    )
    fcst_df = pd.DataFrame(
        {
            "ds": dates,
            "unique_id": fcst.item_id,
            model_name: point_forecast,
        }
    )
    for q in quantiles:
        fcst_df[f"{model_name}-q-{q}"] = fcst.quantile(q)
    return fcst_df


def gluonts_fcsts_to_df(
    fcsts: Iterable[Forecast],
    quantiles: List[float],
    model_name: str,
) -> pd.DataFrame:
    df = []
    for fcst in tqdm(fcsts):
        fcst_df = gluonts_instance_fcst_to_df(fcst, quantiles, model_name)
        df.append(fcst_df)
    return pd.concat(df).reset_index(drop=True)


def run_lag_llama(
    gluonts_dataset: Dataset,
    horizon: int,
    quantiles: List[float],
    models_dir: str,
) -> Tuple[pd.DataFrame, float, str]:
    init_time = time()
    predictor = get_lag_llama_predictor(horizon, models_dir)
    fcsts = predictor.predict(gluonts_dataset, num_samples=100)
    model_name = "LagLlama"
    fcsts_df = gluonts_fcsts_to_df(
        fcsts,
        quantiles=quantiles,
        model_name=model_name,
    )
    total_time = time() - init_time
    return fcsts_df, total_time, model_name


def main(dataset: str):
    exp = ExperimentHandler(dataset)
    fcst_df, total_time, model_name = run_lag_llama(
        gluonts_dataset=exp.gluonts_train_dataset,
        horizon=exp.horizon,
        quantiles=exp.quantiles,
        models_dir=exp.models_dir,
    )
    exp._save_results(fcst_df, total_time, model_name)


if __name__ == "__main__":
    fire.Fire(main)



import httpx
import pytest

import numpy as np
import pandas as pd
import zstandard as zstd

from contextlib import contextmanager
from copy import deepcopy
from nixtla_tests.conftest import HYPER_PARAMS_TEST
from nixtla_tests.helpers.checks import check_num_partitions_same_results
from nixtla_tests.helpers.checks import check_equal_fcsts_add_history


CAPTURED_REQUEST = None


class CapturingClient(httpx.Client):
    def post(self, *args, **kwargs):
        request = self.build_request("POST", *args, **kwargs)
        global CAPTURED_REQUEST
        CAPTURED_REQUEST = {
            "headers": dict(request.headers),
            "content": request.content,
            "method": request.method,
            "url": str(request.url),
        }
        return super().post(*args, **kwargs)


@contextmanager
def capture_request():
    original_client = httpx.Client
    httpx.Client = CapturingClient
    try:
        yield
    finally:
        httpx.Client = original_client


@pytest.mark.parametrize(
    "df_converter, freq",
    [
        pytest.param(lambda series, with_gaps: with_gaps, "5min", id="gaps"),
        pytest.param(
            lambda series, with_gaps: pd.concat([series, series]),
            "5min",
            id="duplicates",
        ),
        pytest.param(lambda series, with_gaps: series, "1min", id="wrong_freq"),
    ],
)
def test_forecast_with_error(series_with_gaps, nixtla_test_client, df_converter, freq):
    series, with_gaps = series_with_gaps

    with pytest.raises(
        ValueError,
        match="missing or duplicate timestamps, or the timestamps do not match",
    ):
        nixtla_test_client.forecast(df=df_converter(series, with_gaps), h=1, freq=freq)


def test_cv_forecast_consistency(nixtla_test_client, cv_series_with_features):
    series_with_features, train, valid, x_cols, h, freq = cv_series_with_features
    for hist_exog_list in [None, [], [x_cols[2], x_cols[1]], x_cols]:
        cv_res = nixtla_test_client.cross_validation(
            series_with_features,
            n_windows=1,
            h=h,
            freq=freq,
            hist_exog_list=hist_exog_list,
        )
        fcst_res = nixtla_test_client.forecast(
            train,
            h=h,
            freq=freq,
            hist_exog_list=hist_exog_list,
            X_df=valid,
        )
        np.testing.assert_allclose(
            cv_res["TimeGPT"], fcst_res["TimeGPT"], atol=1e-4, rtol=1e-3
        )


def test_forecast_different_hist_exog_gives_different_results(
    nixtla_test_client, cv_series_with_features
):
    _, train, valid, x_cols, h, freq = cv_series_with_features
    for X_df in (None, valid):
        res1 = nixtla_test_client.forecast(
            train, h=h, X_df=X_df, freq=freq, hist_exog_list=x_cols[:2]
        )
        res2 = nixtla_test_client.forecast(
            train, h=h, X_df=X_df, freq=freq, hist_exog_list=x_cols[2:]
        )
        with pytest.raises(AssertionError):
            np.testing.assert_allclose(
                res1["TimeGPT"],
                res2["TimeGPT"],
                atol=1e-4,
                rtol=1e-3,
            )


def test_forecast_date_features_multiple_series_and_different_ends(
    nixtla_test_client, two_short_series
):
    h = 12
    fcst_test_series = nixtla_test_client.forecast(
        two_short_series, h=h, date_features=["dayofweek"]
    )
    uids = two_short_series["unique_id"]
    for uid in uids:
        expected = pd.date_range(
            periods=h + 1, start=two_short_series.query("unique_id == @uid")["ds"].max()
        )[1:].tolist()
        actual = fcst_test_series.query("unique_id == @uid")["ds"].tolist()
        assert actual == expected


def test_compression(nixtla_test_client, series_1MB_payload):
    with capture_request():
        nixtla_test_client.forecast(
            df=series_1MB_payload,
            freq="D",
            h=1,
            hist_exog_list=["static_0", "static_1"],
        )
        assert CAPTURED_REQUEST["headers"]["content-encoding"] == "zstd"
        content = CAPTURED_REQUEST["content"]
        assert len(content) < 2**20
        assert len(zstd.ZstdDecompressor().decompress(content)) > 2**20


def test_cv_refit_equivalence(nixtla_test_client, air_passengers_df):
    cv_kwargs = dict(
        df=air_passengers_df,
        n_windows=2,
        h=12,
        freq="MS",
        time_col="timestamp",
        target_col="value",
        finetune_steps=2,
    )
    res_refit = nixtla_test_client.cross_validation(refit=True, **cv_kwargs)
    res_no_refit = nixtla_test_client.cross_validation(refit=False, **cv_kwargs)
    np.testing.assert_allclose(res_refit["value"], res_no_refit["value"])
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(
            res_refit["TimeGPT"],
            res_no_refit["TimeGPT"],
            atol=1e-4,
            rtol=1e-3,
        )


def test_forecast_quantiles_error(nixtla_test_client, air_passengers_df):
    with pytest.raises(Exception) as excinfo:
        nixtla_test_client.forecast(
            df=air_passengers_df,
            h=12,
            time_col="timestamp",
            target_col="value",
            level=[80],
            quantiles=[0.2, 0.3],
        )
    assert "not both" in str(excinfo.value)


@pytest.mark.parametrize(
    "method,kwargs",
    [
        ("forecast", {}),
        ("forecast", {"add_history": True}),
        ("cross_validation", {}),
    ],
)
def test_forecast_quantiles_output(
    nixtla_test_client, air_passengers_df, method, kwargs
):
    test_qls = list(np.arange(0.1, 1, 0.1))
    exp_q_cols = [f"TimeGPT-q-{int(100 * q)}" for q in test_qls]

    args = {
        "df": air_passengers_df,
        "h": 12,
        "time_col": "timestamp",
        "target_col": "value",
        "quantiles": test_qls,
        **kwargs,
    }
    if method == "cross_validation":
        func = nixtla_test_client.cross_validation
    elif method == "forecast":
        func = nixtla_test_client.forecast

    df_qls = func(**args)

    assert all(col in df_qls.columns for col in exp_q_cols)
    assert not any("-lo-" in col for col in df_qls.columns)
    # test monotonicity of quantiles
    for c1, c2 in zip(exp_q_cols[:-1], exp_q_cols[1:]):
        assert df_qls[c1].lt(df_qls[c2]).all()


@pytest.mark.parametrize("freq", ["D", "W-THU", "Q-DEC", "15T"])
@pytest.mark.parametrize(
    "method_name,method_kwargs,exog",
    [
        ("detect_anomalies", {"level": 98}, False),
        ("cross_validation", {"h": 7, "n_windows": 2}, False),
        ("forecast", {"h": 7, "add_history": True}, False),
        ("detect_anomalies", {"level": 98}, True),
        ("cross_validation", {"h": 7, "n_windows": 2}, False),
        ("forecast", {"h": 7, "add_history": True}, False),
    ],
)
def test_num_partitions_same_results_parametrized(
    nixtla_test_client, df_freq_generator, method_name, method_kwargs, freq, exog
):
    mathod_mapper = {
        "detect_anomalies": nixtla_test_client.detect_anomalies,
        "cross_validation": nixtla_test_client.cross_validation,
        "forecast": nixtla_test_client.forecast,
    }
    method = mathod_mapper[method_name]

    df_freq = df_freq_generator(n_series=10, min_length=500, max_length=550, freq=freq)
    df_freq["ds"] = df_freq.groupby("unique_id", observed=True)["ds"].transform(
        lambda x: pd.date_range(periods=len(x), freq=freq, end="2023-01-01")
    )
    if exog:
        df_freq["exog_1"] = 1

    kwargs = {
        "method": method,
        "num_partitions": 2,
        "df": df_freq,
        **method_kwargs,
    }

    check_num_partitions_same_results(**kwargs)


@pytest.mark.parametrize(
    "freq,h",
    [
        ("D", 7),
        ("W-THU", 52),
        ("Q-DEC", 8),
        ("15T", 4 * 24 * 7),
    ],
)
def test_forecast_models_different_results(
    nixtla_test_client, df_freq_generator, freq, h
):
    df_freq = df_freq_generator(n_series=10, min_length=500, max_length=550, freq=freq)
    df_freq["ds"] = df_freq.groupby("unique_id", observed=True)["ds"].transform(
        lambda x: pd.date_range(periods=len(x), freq=freq, end="2023-01-01")
    )
    kwargs = dict(df=df_freq, h=h)
    fcst_1_df = check_equal_fcsts_add_history(
        nixtla_test_client, **{**kwargs, "model": "timegpt-1"}
    )
    fcst_2_df = check_equal_fcsts_add_history(
        nixtla_test_client, **{**kwargs, "model": "timegpt-1-long-horizon"}
    )
    with pytest.raises(
        AssertionError, match=r'\(column name="TimeGPT"\) are different'
    ):
        pd.testing.assert_frame_equal(fcst_1_df, fcst_2_df)


@pytest.mark.parametrize(
    "method, method_kwargs",
    [
        (
            "forecast",
            dict(
                h=12,
                level=[90, 95],
                add_history=True,
                time_col="timestamp",
                target_col="value",
            ),
        ),
        (
            "cross_validation",
            dict(h=12, level=[90, 95], time_col="timestamp", target_col="value"),
        ),
        ("detect_anomalies", dict(level=99, time_col="timestamp", target_col="value")),
    ],
)
def test_different_models_give_different_results(
    air_passengers_df, nixtla_test_client, method, method_kwargs
):
    mathod_mapper = {
        "detect_anomalies": nixtla_test_client.detect_anomalies,
        "cross_validation": nixtla_test_client.cross_validation,
        "forecast": nixtla_test_client.forecast,
    }
    execute = mathod_mapper[method]

    # Run with first model
    out1 = execute(df=air_passengers_df, model="timegpt-1", **method_kwargs)
    # Run with second model
    out2 = execute(
        df=air_passengers_df, model="timegpt-1-long-horizon", **method_kwargs
    )
    # Compare only the TimeGPT column
    with pytest.raises(
        AssertionError, match=r'\(column name="TimeGPT"\) are different'
    ):
        pd.testing.assert_frame_equal(out1[["TimeGPT"]], out2[["TimeGPT"]])

    # test unsupported model
    method_kwargs["model"] = "my-awesome-model"
    with pytest.raises(ValueError, match="unsupported model"):
        execute(df=air_passengers_df, **method_kwargs)


def test_shap_features(nixtla_test_client, date_features_result):
    # Test shap values are returned and sum to predictions
    df_date_features, future_df, _ = date_features_result
    h = 12
    fcst_df = nixtla_test_client.forecast(
        df=df_date_features, h=h, X_df=future_df, feature_contributions=True
    )
    shap_values = nixtla_test_client.feature_contributions
    assert len(shap_values) == len(fcst_df)
    np.testing.assert_allclose(
        fcst_df["TimeGPT"].values, shap_values.iloc[:, 3:].sum(axis=1).values, rtol=1e-3
    )

    fcst_hist_df = nixtla_test_client.forecast(
        df=df_date_features,
        h=h,
        X_df=future_df,
        add_history=True,
        feature_contributions=True,
    )
    shap_values_hist = nixtla_test_client.feature_contributions
    assert len(shap_values_hist) == len(fcst_hist_df)
    np.testing.assert_allclose(
        fcst_hist_df["TimeGPT"].values,
        shap_values_hist.iloc[:, 3:].sum(axis=1).values,
        atol=1e-4,
    )

    # test num partitions
    _ = nixtla_test_client.feature_contributions
    pd.testing.assert_frame_equal(
        nixtla_test_client.feature_contributions, shap_values_hist, atol=1e-4, rtol=1e-3
    )


@pytest.mark.parametrize("hyp", HYPER_PARAMS_TEST)
def test_exogenous_variables_cv(nixtla_test_client, exog_data, hyp):
    df_ex_, df_train, df_test, x_df_test = exog_data
    fcst_test = nixtla_test_client.forecast(
        df_train.merge(df_ex_.drop(columns="y")), h=12, X_df=x_df_test, **hyp
    )
    fcst_test = df_test[["unique_id", "ds", "y"]].merge(fcst_test)
    fcst_test = fcst_test.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    fcst_cv = nixtla_test_client.cross_validation(df_ex_, h=12, **hyp)
    fcst_cv = fcst_cv.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    pd.testing.assert_frame_equal(
        fcst_test,
        fcst_cv.drop(columns="cutoff"),
        atol=1e-4,
        rtol=1e-3,
    )


@pytest.mark.parametrize("hyp", HYPER_PARAMS_TEST)
def test_forecast_vs_cv_no_exog(
    nixtla_test_client, train_test_split, air_passengers_renamed_df, hyp
):
    df_train, df_test = train_test_split
    fcst_test = nixtla_test_client.forecast(df_train, h=12, **hyp)
    fcst_test = df_test[["unique_id", "ds", "y"]].merge(fcst_test)
    fcst_test = fcst_test.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    fcst_cv = nixtla_test_client.cross_validation(
        air_passengers_renamed_df, h=12, **hyp
    )
    fcst_cv = fcst_cv.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    pd.testing.assert_frame_equal(
        fcst_test,
        fcst_cv.drop(columns="cutoff"),
        rtol=1e-2,
    )


@pytest.mark.parametrize("hyp", HYPER_PARAMS_TEST)
def test_forecast_vs_cv_insert_y(
    nixtla_test_client, train_test_split, air_passengers_renamed_df, hyp
):
    df_train, df_test = train_test_split
    fcst_test = nixtla_test_client.forecast(df_train, h=12, **hyp)
    fcst_test.insert(2, "y", df_test["y"].values)
    fcst_test = fcst_test.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    fcst_cv = nixtla_test_client.cross_validation(
        air_passengers_renamed_df, h=12, **hyp
    )
    fcst_cv = fcst_cv.sort_values(["unique_id", "ds"]).reset_index(drop=True)
    pd.testing.assert_frame_equal(
        fcst_test,
        fcst_cv.drop(columns="cutoff"),
        rtol=1e-2,
    )


def test_forecast_and_anomalies_index_vs_columns(
    nixtla_test_client, air_passengers_renamed_df, air_passengers_renamed_df_with_index
):
    fcst_inferred_df_index = nixtla_test_client.forecast(
        air_passengers_renamed_df_with_index, h=10
    )
    anom_inferred_df_index = nixtla_test_client.detect_anomalies(
        air_passengers_renamed_df_with_index
    )
    fcst_inferred_df = nixtla_test_client.forecast(
        air_passengers_renamed_df[["ds", "unique_id", "y"]], h=10
    )
    anom_inferred_df = nixtla_test_client.detect_anomalies(
        air_passengers_renamed_df[["ds", "unique_id", "y"]]
    )
    pd.testing.assert_frame_equal(
        fcst_inferred_df_index, fcst_inferred_df, atol=1e-4, rtol=1e-3
    )
    pd.testing.assert_frame_equal(
        anom_inferred_df_index, anom_inferred_df, atol=1e-4, rtol=1e-3
    )


@pytest.mark.parametrize("freq", ["Y", "W-MON", "Q-DEC", "H"])
def test_forecast_index_vs_columns_various_freq(
    nixtla_test_client, air_passengers_renamed_df_with_index, freq
):
    df_ds_index = air_passengers_renamed_df_with_index.groupby("unique_id").tail(80)
    df_ds_index.index = np.concatenate(
        df_ds_index["unique_id"].nunique()
        * [pd.date_range(end="2023-01-01", periods=80, freq=freq)]
    )
    df_ds_index.index.name = "ds"
    fcst_inferred_df_index = nixtla_test_client.forecast(df_ds_index, h=10)
    df_test = df_ds_index.reset_index()
    fcst_inferred_df = nixtla_test_client.forecast(df_test, h=10)
    pd.testing.assert_frame_equal(
        fcst_inferred_df_index, fcst_inferred_df, atol=1e-4, rtol=1e-3
    )


def test_index_as_time_col(nixtla_test_client, air_passengers_df):
    df_test = deepcopy(air_passengers_df)
    df_test["timestamp"] = pd.to_datetime(df_test["timestamp"])
    df_test.set_index(df_test["timestamp"], inplace=True)
    df_test.drop(columns="timestamp", inplace=True)
    # Using user_provided time_col and freq
    timegpt_anomalies_df_1 = nixtla_test_client.detect_anomalies(
        air_passengers_df, time_col="timestamp", target_col="value", freq="M"
    )
    # Infer time_col and freq from index
    timegpt_anomalies_df_2 = nixtla_test_client.detect_anomalies(
        df_test, time_col="timestamp", target_col="value"
    )
    pd.testing.assert_frame_equal(
        timegpt_anomalies_df_1,
        timegpt_anomalies_df_2,
        atol=1e-4,
        rtol=1e-3,
    )



from time import time


import numpy as np
import pandas as pd
from dotenv import load_dotenv
from nixtla import NixtlaClient

load_dotenv()


def read_and_prepare_data(file_path: str, value_name: str = "y") -> pd.DataFrame:
    """Reads data in wide format, and returns it in long format with columns `unique_id`, `ds`, `y`"""
    df = pd.read_csv(file_path)
    uid_cols = ["Client", "Warehouse", "Product"]
    df["unique_id"] = df[uid_cols].astype(str).agg("-".join, axis=1)
    df = df.drop(uid_cols, axis=1)
    df = df.melt(id_vars=["unique_id"], var_name="ds", value_name=value_name)
    df["ds"] = pd.to_datetime(df["ds"])
    df = df.sort_values(by=["unique_id", "ds"])
    return df


def get_train_data() -> pd.DataFrame:
    """Reads all train data and returns it in long format with columns `unique_id`, `ds`, `y`"""
    train_list = [read_and_prepare_data(f"./data/phase_{i}_sales.csv") for i in [0, 1]]
    train_df = pd.concat(train_list).reset_index(drop=True)
    train_df = train_df.sort_values(by=["unique_id", "ds"])

    def remove_leading_zeros(group):
        first_non_zero_index = group["y"].ne(0).idxmax()
        return group.loc[first_non_zero_index:]

    train_df = (
        train_df.groupby("unique_id").apply(remove_leading_zeros).reset_index(drop=True)
    )
    return train_df


def get_competition_forecasts() -> pd.DataFrame:
    """Reads all competition forecasts and returns it in long format with columns `unique_id`, `ds`, `y`"""
    fcst_df: pd.DataFrame | None = None
    for place in ["1st", "2nd", "3rd", "4th", "5th"]:
        fcst_df_place = read_and_prepare_data(
            f"./data/solution_{place}_place.csv", place
        )
        if fcst_df is None:
            fcst_df = fcst_df_place
        else:
            fcst_df = fcst_df.merge(
                fcst_df_place,
                on=["unique_id", "ds"],
                how="left",
            )
    return fcst_df


def vn1_competition_evaluation(forecasts: pd.DataFrame) -> pd.DataFrame:
    """Computes competition evaluation scores"""
    actual = read_and_prepare_data("./data/phase_2_sales.csv")
    res = actual[["unique_id", "ds", "y"]].merge(
        forecasts, on=["unique_id", "ds"], how="left"
    )
    ids_forecasts = forecasts["unique_id"].unique()
    ids_res = res["unique_id"].unique()
    assert set(ids_forecasts) == set(ids_res), "Some unique_ids are missing"
    scores = {}
    for model in [col for col in forecasts.columns if col not in ["unique_id", "ds"]]:
        abs_err = np.nansum(np.abs(res[model] - res["y"]))
        err = np.nansum(res[model] - res["y"])
        score = abs_err + abs(err)
        score = score / res["y"].sum()
        scores[model] = round(score, 4)
    score_df = pd.DataFrame(list(scores.items()), columns=["model", "score"])
    score_df = score_df.sort_values(by="score")
    return score_df


def main():
    """Complete pipeline"""
    train_df = get_train_data()
    client = NixtlaClient()
    init = time()
    fcst_df = client.forecast(train_df, h=13, model="timegpt-1-long-horizon")
    print(f"TimeGPT time: {time() - init}")
    fcst_df_comp = get_competition_forecasts()
    fcst_df = fcst_df.merge(fcst_df_comp, on=["unique_id", "ds"], how="left")
    eval_df = vn1_competition_evaluation(fcst_df)
    print(eval_df)


if __name__ == "__main__":
    main()



def s3_file_exists(s3, bucket: str, key: str) -> bool:
    from botocore.exceptions import ClientError

    try:
        s3.head_object(Bucket=bucket, Key=key)
        return True
    except ClientError as e:
        if e.response["Error"]["Code"] != "404":
            raise e
    return False


def read_s3_file(s3, bucket: str, key: str) -> str:
    resp = s3.get_object(Bucket=bucket, Key=key)
    return resp["Body"].read().decode("utf-8")


def forecast_partition(
    i: int,
    version: str,
    bucket: str,
    prefix: str,
    n_series: int,
    freq: str = "D",
    h: int = 7,
) -> None:
    import logging
    import math
    import time

    import boto3
    import pandas as pd
    from nixtla import NixtlaClient as V2Client
    from nixtlats import NixtlaClient as V1Client
    from tqdm.auto import tqdm
    from utilsforecast.data import generate_series

    s3 = boto3.client("s3")
    # only process if we haven't saved the time
    time_key = f"{prefix}/times/{version}/{i}.txt"
    if s3_file_exists(s3, bucket, time_key):
        print(f"{i}-th partition already processed, skipping.")
        return
    logging.getLogger("nixtla").setLevel(logging.ERROR)
    logging.getLogger("nixtlats").setLevel(logging.ERROR)

    series = generate_series(
        n_series=n_series,
        freq=freq,
        min_length=100,
        max_length=200,
        seed=i,
    )
    series["unique_id"] = series["unique_id"].astype("uint32") + i * n_series

    start = time.perf_counter()
    if version == "v1":
        client = V1Client()
        # v1 is slower when partitioning, so we do this sequentially
        num_partitions = math.ceil(n_series / 50_000)
        uids = series["unique_id"].unique()
        n_ids = uids.size
        ids_per_part = math.ceil(n_ids / num_partitions)
        results = []
        for j in tqdm(range(0, n_ids, ids_per_part)):
            part_uids = uids[j : j + ids_per_part]
            part = series[series["unique_id"].isin(part_uids)]
            results.append(client.forecast(df=part, h=h, freq=freq))
        forecast = pd.concat(results)
    else:
        client = V2Client()
        num_partitions = math.ceil(n_series / 100_000)
        forecast = client.forecast(
            df=series, h=h, freq=freq, num_partitions=num_partitions
        )
    time_taken = "{:.2f}".format(time.perf_counter() - start)
    forecast.to_parquet(
        f"s3://{bucket}/{prefix}/output/{version}/{i}.parquet", index=False
    )
    s3.put_object(Bucket=bucket, Key=time_key, Body=time_taken)
    print(f"{i}: {time_taken}")


def generate_forecasts(
    version: str,
    bucket: str,
    prefix: str,
    n_partitions: int,
    series_per_partition: int,
    n_jobs: int,
) -> None:
    from concurrent.futures import ProcessPoolExecutor
    from functools import partial

    fn = partial(
        forecast_partition,
        version=version,
        bucket=bucket,
        prefix=prefix,
        n_series=series_per_partition,
    )
    with ProcessPoolExecutor(n_jobs) as executor:
        _ = executor.map(fn, range(n_partitions))


def read_times(
    s3, version: str, bucket: str, prefix: str, n_partitions: int
) -> list[str]:
    from concurrent.futures import ThreadPoolExecutor, as_completed

    from tqdm.auto import tqdm

    key = f"{prefix}/{version}_times.txt"
    if s3_file_exists(s3, bucket, key):
        return read_s3_file(s3, bucket, key).splitlines()
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(
                read_s3_file, s3, bucket, f"{prefix}/times/{version}/{i}.txt"
            )
            for i in range(n_partitions)
        ]
        times = []
        for future in tqdm(as_completed(futures), total=len(futures)):
            times.append(future.result())
    s3.put_object(
        Bucket=bucket,
        Key=key,
        Body="\n".join(times),
    )
    return times


def main(
    bucket: str = "datasets-nixtla",
    prefix: str = "one-billion",
    n_partitions: int = 1_000,
    series_per_partition: int = 1_000_000,
    n_jobs: int = 5,
):
    import boto3
    import pandas as pd

    times = {}
    s3 = boto3.client("s3")
    for version in ("v1", "v2"):
        generate_forecasts(
            version=version,
            bucket=bucket,
            prefix=prefix,
            n_partitions=n_partitions,
            series_per_partition=series_per_partition,
            n_jobs=n_jobs,
        )
        times[version] = read_times(
            s3, version=version, bucket=bucket, prefix=prefix, n_partitions=n_partitions
        )
    pd.DataFrame(times).to_csv(f"s3://{bucket}/{prefix}/times.csv", index=False)


if __name__ == "__main__":
    import fire

    fire.Fire(main)



from pathlib import Path

import fire
from numpy import column_stack
import pandas as pd


def read_kind_results(kind: str, dir: str):
    files = list(Path(dir).rglob(f"*{kind}.parquet"))
    df = pd.concat(
        [pd.read_parquet(file).assign(file=str(file).split("/")[-2]) for file in files],
        ignore_index=True,
    )
    return df


def summarize_results_per_file(metrics_df: pd.DataFrame):
    metrics_df_per_freq = metrics_df.groupby(["file", "metric", "model"]).mean(
        numeric_only=True
    )
    metrics_df_per_freq = metrics_df_per_freq.reset_index()
    metrics_df_per_freq = metrics_df_per_freq.query(
        "model in ['Prophet', 'SeasonalNaive', 'TimeGPT']"
    )
    models = metrics_df_per_freq["model"].unique()
    metrics_df_per_freq = pd.pivot(
        metrics_df_per_freq,
        index=["file", "metric"],
        columns="model",
        values="value",
    ).reset_index()
    for model in models:
        if model == "SeasonalNaive":
            continue
        metrics_df_per_freq[model] /= metrics_df_per_freq["SeasonalNaive"]
    metrics_df_per_freq["SeasonalNaive"] /= metrics_df_per_freq["SeasonalNaive"]
    return metrics_df_per_freq


def prepare_results(df: pd.DataFrame):
    def bold_best(row):
        row = row.round(3)
        models = row.drop(columns=["file", "metric"]).columns
        best_model = row[models].idxmin(axis=1).item()
        row[best_model] = "**" + str(row[best_model].item()) + "**"
        return row

    df_bolded = df.groupby(["file", "metric"]).apply(bold_best)
    df_bolded = df_bolded.reset_index(drop=True)
    return df_bolded


def write_to_readme(content: str):
    with open("README.md", "r") as file:
        readme_content = file.readlines()
    start_index = -1
    end_index = -1
    for i, line in enumerate(readme_content):
        if line.strip().lower() == "## results":
            start_index = i + 1
        if start_index != -1 and line.strip() == "<end>":
            end_index = i
            break

    if start_index != -1 and end_index != -1:
        readme_content = (
            readme_content[: start_index + 1]
            + [content + "\n"]
            + readme_content[end_index:]
        )
    else:
        print("Results section not found or improperly formatted")

    # Write the changes back to the README
    with open("README.md", "w") as file:
        file.writelines(readme_content)


def summarize_results(dir: str):
    metrics_df = read_kind_results("metrics", dir)
    summary_df = read_kind_results("summary", dir)
    summary_df = (
        summary_df.set_index(["file", "frequency"])
        .reset_index()
        .round(3)
        .sort_values("frequency")
    )
    no_int_cols = ["file", "frequency", "mean", "std"]
    for col in summary_df.columns:
        if col not in no_int_cols:
            summary_df[col] = summary_df[col].astype(int)
    summary_df = summary_df.to_markdown(index=False, intfmt=",", floatfmt=",.3f")
    time_df = read_kind_results("time", dir)
    time_df = time_df.assign(metric="time").rename(columns={"time": "value"})
    metrics_df_per_file = summarize_results_per_file(metrics_df)
    time_df = summarize_results_per_file(time_df)
    eval_df = pd.concat([metrics_df_per_file, time_df], ignore_index=True)
    eval_df = prepare_results(eval_df)[
        ["file", "metric", "TimeGPT", "Prophet", "SeasonalNaive"]
    ]
    n_files = eval_df["file"].nunique()
    eval_df = eval_df.to_markdown(
        index=False,
        colalign=2 * ["left"] + (eval_df.shape[1] - 2) * ["right"],
    )
    markdown_lines = eval_df.split("\n")
    custom_separator = markdown_lines[1].replace(":", "-")
    for i in range(4, len(markdown_lines) + n_files - 1, 4):
        markdown_lines.insert(i + 1, custom_separator)
    markdown_lines.insert(
        0,
        ("\n### Data Description\n\n" f"{summary_df}\n\n" "### Performance\n\n"),
    )
    eval_df = "\n".join(markdown_lines)
    write_to_readme(eval_df)


if __name__ == "__main__":
    fire.Fire(summarize_results)



from typing import List

import pandas as pd
from gluonts.time_feature.seasonality import get_seasonality as _get_seasonality
from tqdm import tqdm
from utilsforecast.processing import (
    backtest_splits,
    drop_index_if_pandas,
    join,
    maybe_compute_sort_indices,
    take_rows,
    vertical_concat,
)


def get_seasonality(freq: str) -> int:
    return _get_seasonality(freq, seasonalities={"D": 7})


def maybe_convert_col_to_datetime(df: pd.DataFrame, col_name: str) -> pd.DataFrame:
    if not pd.api.types.is_datetime64_any_dtype(df[col_name]):
        df = df.copy()
        df[col_name] = pd.to_datetime(df[col_name])
    return df


class Forecaster:
    def forecast(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
    ) -> pd.DataFrame:
        raise NotImplementedError

    def cross_validation(
        self,
        df: pd.DataFrame,
        h: int,
        freq: str,
        n_windows: int = 1,
        step_size: int | None = None,
    ) -> pd.DataFrame:
        df = maybe_convert_col_to_datetime(df, "ds")
        # mlforecast cv code
        results = []
        sort_idxs = maybe_compute_sort_indices(df, "unique_id", "ds")
        if sort_idxs is not None:
            df = take_rows(df, sort_idxs)
        splits = backtest_splits(
            df,
            n_windows=n_windows,
            h=h,
            id_col="unique_id",
            time_col="ds",
            freq=pd.tseries.frequencies.to_offset(freq),
            step_size=h if step_size is None else step_size,
        )
        for _, (cutoffs, train, valid) in tqdm(enumerate(splits)):
            if len(valid.columns) > 3:
                raise NotImplementedError(
                    "Cross validation with exogenous variables is not yet supported."
                )
            y_pred = self.forecast(
                df=train,
                h=h,
                freq=freq,
            )
            y_pred = join(y_pred, cutoffs, on="unique_id", how="left")
            result = join(
                valid[["unique_id", "ds", "y"]],
                y_pred,
                on=["unique_id", "ds"],
            )
            if result.shape[0] < valid.shape[0]:
                raise ValueError(
                    "Cross validation result produced less results than expected. "
                    "Please verify that the frequency parameter (freq) matches your series' "
                    "and that there aren't any missing periods."
                )
            results.append(result)
        out = vertical_concat(results)
        out = drop_index_if_pandas(out)
        first_out_cols = ["unique_id", "ds", "cutoff", "y"]
        remaining_cols = [c for c in out.columns if c not in first_out_cols]
        fcst_cv_df = out[first_out_cols + remaining_cols]
        return fcst_cv_df



import pandas as pd
import pytest
from utilsforecast.data import generate_series

from .utils import models


@pytest.mark.parametrize("model", models)
@pytest.mark.parametrize("freq", ["H", "D", "W-MON", "MS"])
@pytest.mark.parametrize("h", [1, 12])
def test_correct_forecast_dates(model, freq, h):
    n_series = 5
    df = generate_series(
        n_series,
        freq=freq,
    )
    df["unique_id"] = df["unique_id"].astype(str)
    df_test = df.groupby("unique_id").tail(h)
    df_train = df.drop(df_test.index)
    fcst_df = model.forecast(
        df_train,
        h=h,
        freq=freq,
    )
    exp_n_cols = 3
    assert fcst_df.shape == (n_series * h, exp_n_cols)
    exp_cols = ["unique_id", "ds"]
    pd.testing.assert_frame_equal(
        fcst_df[exp_cols].sort_values(["unique_id", "ds"]).reset_index(drop=True),
        df_test[exp_cols].sort_values(["unique_id", "ds"]).reset_index(drop=True),
    )


@pytest.mark.parametrize("model", models)
@pytest.mark.parametrize("freq", ["H", "D", "W-MON", "MS"])
@pytest.mark.parametrize("n_windows", [1, 4])
def test_cross_validation(model, freq, n_windows):
    h = 12
    n_series = 5
    df = generate_series(n_series, freq=freq, equal_ends=True)
    df["unique_id"] = df["unique_id"].astype(str)
    cv_df = model.cross_validation(
        df,
        h=h,
        freq=freq,
        n_windows=n_windows,
    )
    exp_n_cols = 5  # unique_id, cutoff, ds, y, model
    assert cv_df.shape == (n_series * h * n_windows, exp_n_cols)
    cutoffs = cv_df["cutoff"].unique()
    assert len(cutoffs) == n_windows
    df_test = df.groupby("unique_id").tail(h * n_windows)
    exp_cols = ["unique_id", "ds", "y"]
    pd.testing.assert_frame_equal(
        cv_df.sort_values(["unique_id", "ds"]).reset_index(drop=True)[exp_cols],
        df_test.sort_values(["unique_id", "ds"]).reset_index(drop=True)[exp_cols],
    )
    if n_windows == 1:
        # test same results using predict with less data
        df_test = df.groupby("unique_id").tail(h)
        df_train = df.drop(df_test.index)
        fcst_df = model.forecast(
            df_train,
            h=h,
            freq=freq,
        )
        exp_cols = ["unique_id", "ds"]
        pd.testing.assert_frame_equal(
            cv_df.sort_values(["unique_id", "ds"]).reset_index(drop=True)[exp_cols],
            fcst_df.sort_values(["unique_id", "ds"]).reset_index(drop=True)[exp_cols],
        )
